# Copyright 2026 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import operator

import jax
import jax.numpy as jnp
import numpy as np
import pytest

import brainevent
from brainevent import DataRepresentation
from brainevent._data import JITCMatrix, _initialize_seed, _initialize_conn_length


class _DummyJITCMatrix(JITCMatrix):
    def __init__(self):
        super().__init__((), shape=(0, 0))

    def transpose(self, axes=None):
        return self

    def tree_flatten(self):
        return (), None

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls()

    def _unitary_op(self, op):
        return ("unitary", op)

    def _binary_op(self, other, op):
        return ("binary", op, other)

    def _binary_rop(self, other, op):
        return ("binary_r", op, other)


def test_unitary_operator_dispatch():
    mat = _DummyJITCMatrix()
    assert mat.__abs__() == ("unitary", operator.abs)
    assert mat.__neg__() == ("unitary", operator.neg)
    assert mat.__pos__() == ("unitary", operator.pos)


def test_binary_operator_dispatch():
    mat = _DummyJITCMatrix()
    other = jnp.asarray(2.0)
    assert mat * other == ("binary", operator.mul, other)
    assert mat / other == ("binary", operator.truediv, other)
    assert mat.__truediv__(other) == ("binary", operator.truediv, other)
    assert mat + other == ("binary", operator.add, other)
    assert mat - other == ("binary", operator.sub, other)
    assert mat % other == ("binary", operator.mod, other)


def test_binary_reflected_operator_dispatch():
    mat = _DummyJITCMatrix()
    other = jnp.asarray(3.0)
    assert other * mat == ("binary_r", operator.mul, other)
    assert other / mat == ("binary_r", operator.truediv, other)
    assert mat.__rtruediv__(other) == ("binary_r", operator.truediv, other)
    assert other + mat == ("binary_r", operator.add, other)
    assert other - mat == ("binary_r", operator.sub, other)
    assert other % mat == ("binary_r", operator.mod, other)


def test_initialize_seed_explicit_and_array():
    seed = _initialize_seed(123)
    assert seed.shape == (1,)
    assert seed.dtype == jnp.int32
    assert int(seed[0]) == 123

    seed_arr = _initialize_seed(np.asarray([5, 7], dtype=np.int64))
    assert seed_arr.shape == (2,)
    assert seed_arr.dtype == jnp.int32
    assert np.array_equal(np.asarray(seed_arr), np.asarray([5, 7], dtype=np.int32))


def test_initialize_seed_none():
    seed = _initialize_seed(None)
    assert seed.shape == (1,)
    assert seed.dtype == jnp.int32
    value = int(seed[0])
    assert 0 <= value < int(1e8)


def test_initialize_conn_length_values():
    clen = _initialize_conn_length(0.25)
    assert clen.dtype == jnp.int32
    assert int(clen) == 8

    clen2 = _initialize_conn_length(0.6)
    assert clen2.dtype == jnp.int32
    assert int(clen2) == 4


def test_initialize_conn_length_jit():
    @jax.jit
    def f(p):
        return _initialize_conn_length(p)

    clen = f(0.5)
    assert int(clen) == 4


# ---------------------------------------------------------------------------
# Helpers: minimal concrete subclass for testing the base mechanism
# ---------------------------------------------------------------------------

@jax.tree_util.register_pytree_node_class
class _SimpleBuffered(DataRepresentation):
    """Minimal concrete DataRepresentation for testing the buffer API."""

    def __init__(self, value, *, shape, buffers=None):
        self.value = jnp.asarray(value)
        super().__init__((value,), shape=shape)  # creates _buffer_registry
        self.register_buffer('cached_sum', None)
        self.register_buffer('label', None)
        if buffers is not None:
            for name, val in buffers.items():
                self.register_buffer(name, val)

    def transpose(self, axes=None):
        return self

    def tree_flatten(self):
        aux = {'shape': self.shape, 'value': self.value}
        return (), (aux, self.buffers)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        obj = object.__new__(cls)
        aux_data, buffer = aux_data
        obj._buffer_registry = set(buffer.keys())
        for k, v in aux_data.items():
            setattr(obj, k, v)
        for k, v in buffer.items():
            setattr(obj, k, v)
        return obj


# ===========================================================================
# 1. register_buffer basics
# ===========================================================================

class TestRegisterBuffer:
    def test_register_with_default_none(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        assert obj.cached_sum is None
        assert obj.label is None

    def test_register_with_value(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        obj.register_buffer('extra', 42)
        assert obj.extra == 42
        assert 'extra' in obj._buffer_registry

    def test_registry_tracks_names(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        assert 'cached_sum' in obj._buffer_registry
        assert 'label' in obj._buffer_registry
        assert len(obj._buffer_registry) == 2

    def test_register_overwrites_previous_value(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        obj.register_buffer('cached_sum', 10)
        assert obj.cached_sum == 10
        obj.register_buffer('cached_sum', 20)
        assert obj.cached_sum == 20
        assert list(obj._buffer_registry).count('cached_sum') == 1


# ===========================================================================
# 2. set_buffer
# ===========================================================================

class TestSetBuffer:
    def test_set_registered_buffer(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        obj.set_buffer('cached_sum', jnp.array(99.0))
        assert float(obj.cached_sum) == 99.0

    def test_set_unregistered_raises(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        with pytest.raises(ValueError, match="not registered"):
            obj.set_buffer('nonexistent', 5)


# ===========================================================================
# 3. buffers property
# ===========================================================================

class TestBuffersProperty:
    def test_returns_dict_of_registered_buffers(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        bufs = obj.buffers
        assert isinstance(bufs, dict)
        assert set(bufs.keys()) == {'cached_sum', 'label'}
        assert bufs['cached_sum'] is None
        assert bufs['label'] is None

    def test_reflects_updated_values(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        obj.cached_sum = jnp.array(3.14)
        obj.label = 'test'
        bufs = obj.buffers
        assert float(bufs['cached_sum']) == pytest.approx(3.14)
        assert bufs['label'] == 'test'

    def test_buffers_returns_new_dict_each_time(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        d1 = obj.buffers
        d2 = obj.buffers
        assert d1 == d2
        assert d1 is not d2


# ===========================================================================
# 4. Constructor buffers kwarg
# ===========================================================================

class TestBuffersKwarg:
    def test_overrides_defaults(self):
        bufs = {'cached_sum': jnp.array(7.0), 'label': 'hello'}
        obj = _SimpleBuffered(1.0, shape=(2, 2), buffers=bufs)
        assert float(obj.cached_sum) == 7.0
        assert obj.label == 'hello'

    def test_partial_override(self):
        bufs = {'cached_sum': jnp.array(5.0)}
        obj = _SimpleBuffered(1.0, shape=(2, 2), buffers=bufs)
        assert float(obj.cached_sum) == 5.0
        assert obj.label is None

    def test_none_is_noop(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2), buffers=None)
        assert obj.cached_sum is None
        assert obj.label is None

    def test_empty_dict_is_noop(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2), buffers={})
        assert obj.cached_sum is None
        assert obj.label is None

    def test_registers_new_buffers(self):
        bufs = {'cached_sum': 1, 'label': 2, 'brand_new': 42}
        obj = _SimpleBuffered(1.0, shape=(2, 2), buffers=bufs)
        assert obj.brand_new == 42
        assert 'brand_new' in obj._buffer_registry


# ===========================================================================
# 5. JAX pytree round-trip
# ===========================================================================

class TestPytreeRoundTrip:
    def test_flatten_unflatten_preserves_buffers(self):
        obj = _SimpleBuffered(1.0, shape=(3, 3))
        obj.set_buffer('cached_sum', jnp.array(123.0))
        obj.set_buffer('label', 'test_label')

        children, aux = obj.tree_flatten()
        restored = _SimpleBuffered.tree_unflatten(aux, children)

        assert float(restored.cached_sum) == 123.0
        assert restored.label == 'test_label'
        assert restored._buffer_registry == {'cached_sum', 'label'}

    def test_jax_tree_map_preserves_buffers(self):
        obj = _SimpleBuffered(1.0, shape=(3, 3))
        obj.set_buffer('cached_sum', jnp.array(5.0))

        mapped = jax.tree.map(lambda x: x, obj)
        assert isinstance(mapped, _SimpleBuffered)
        assert float(mapped.cached_sum) == 5.0
        assert mapped._buffer_registry == {'cached_sum', 'label'}

    def test_jit_preserves_buffers(self):
        obj = _SimpleBuffered(jnp.array(2.0), shape=(3, 3))
        obj.set_buffer('cached_sum', jnp.array(10.0))

        @jax.jit
        def identity(x):
            return x

        result = identity(obj)
        assert isinstance(result, _SimpleBuffered)
        assert float(result.cached_sum) == 10.0

    def test_none_buffer_values_roundtrip(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        children, aux = obj.tree_flatten()
        restored = _SimpleBuffered.tree_unflatten(aux, children)
        assert restored.cached_sum is None
        assert restored.label is None
        assert restored._buffer_registry == {'cached_sum', 'label'}

    def test_registry_is_mutable_set_after_unflatten(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        children, aux = obj.tree_flatten()
        restored = _SimpleBuffered.tree_unflatten(aux, children)
        assert isinstance(restored._buffer_registry, set)
        restored.register_buffer('dynamic', 99)
        assert 'dynamic' in restored._buffer_registry

    def test_repeated_unflatten_is_safe(self):
        """tree_unflatten must not mutate shared aux_data (JAX may call it multiple times)."""
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        obj.set_buffer('cached_sum', jnp.array(42.0))

        children, aux = obj.tree_flatten()
        r1 = _SimpleBuffered.tree_unflatten(aux, children)
        r2 = _SimpleBuffered.tree_unflatten(aux, children)
        assert r1._buffer_registry == {'cached_sum', 'label'}
        assert r2._buffer_registry == {'cached_sum', 'label'}
        assert float(r2.cached_sum) == 42.0


# ===========================================================================
# 6. CSR buffer integration
# ===========================================================================

class TestCSRBuffers:
    @pytest.fixture
    def csr_mat(self):
        data = jnp.array([1.0, 2.0, 3.0, 4.0])
        indices = jnp.array([0, 1, 0, 2])
        indptr = jnp.array([0, 2, 4])
        return brainevent.CSR((data, indices, indptr), shape=(2, 3))

    def test_csr_no_buffers_initially(self, csr_mat):
        assert csr_mat.buffers == {}
        assert not hasattr(csr_mat, 'diag_positions')

    def test_csr_register_buffer_then_access(self, csr_mat):
        csr_mat.register_buffer('diag_positions', jnp.array([0, 3]))
        assert 'diag_positions' in csr_mat._buffer_registry
        np.testing.assert_array_equal(csr_mat.diag_positions, jnp.array([0, 3]))

    def test_csr_pytree_roundtrip(self, csr_mat):
        csr_mat.register_buffer('diag_positions', jnp.array([0, 3]))
        children, aux = csr_mat.tree_flatten()
        restored = brainevent.CSR.tree_unflatten(aux, children)
        np.testing.assert_array_equal(restored.diag_positions, jnp.array([0, 3]))
        assert 'diag_positions' in restored._buffer_registry

    def test_csr_jit_preserves_registered_buffer(self, csr_mat):
        csr_mat.register_buffer('diag_positions', jnp.array([0, 3]))

        @jax.jit
        def identity(m):
            return m

        result = identity(csr_mat)
        np.testing.assert_array_equal(result.diag_positions, jnp.array([0, 3]))

    def test_csr_with_data_preserves_buffers(self, csr_mat):
        csr_mat.register_buffer('diag_positions', jnp.array([0, 3]))
        new_data = jnp.array([10.0, 20.0, 30.0, 40.0])
        new_csr = csr_mat.with_data(new_data)
        np.testing.assert_array_equal(new_csr.diag_positions, jnp.array([0, 3]))
        np.testing.assert_array_equal(new_csr.data, new_data)

    def test_csr_apply_preserves_buffers(self, csr_mat):
        csr_mat.register_buffer('diag_positions', jnp.array([0, 3]))
        squared = csr_mat.apply(lambda x: x ** 2)
        np.testing.assert_array_equal(squared.diag_positions, jnp.array([0, 3]))
        np.testing.assert_array_equal(squared.data, jnp.array([1.0, 4.0, 9.0, 16.0]))

    def test_csr_transpose_preserves_buffers(self, csr_mat):
        csr_mat.register_buffer('diag_positions', jnp.array([0, 3]))
        csc = csr_mat.transpose()
        assert 'diag_positions' in csc._buffer_registry
        np.testing.assert_array_equal(csc.diag_positions, jnp.array([0, 3]))

    def test_csr_arithmetic_preserves_buffers(self, csr_mat):
        csr_mat.register_buffer('diag_positions', jnp.array([0, 3]))
        scaled = csr_mat * 2.0
        assert 'diag_positions' in scaled._buffer_registry
        np.testing.assert_array_equal(scaled.diag_positions, jnp.array([0, 3]))

    def test_csr_reflected_arithmetic_preserves_buffers(self, csr_mat):
        csr_mat.register_buffer('diag_positions', jnp.array([0, 3]))
        scaled = 2.0 * csr_mat
        assert 'diag_positions' in scaled._buffer_registry
        np.testing.assert_array_equal(scaled.diag_positions, jnp.array([0, 3]))

    def test_csr_buffers_kwarg_in_constructor(self):
        data = jnp.array([1.0, 2.0])
        indices = jnp.array([0, 1])
        indptr = jnp.array([0, 1, 2])
        bufs = {'diag_positions': jnp.array([0, 1])}
        csr = brainevent.CSR((data, indices, indptr), shape=(2, 2), buffers=bufs)
        np.testing.assert_array_equal(csr.diag_positions, jnp.array([0, 1]))
        assert 'diag_positions' in csr._buffer_registry

    def test_csr_no_buffers_pytree_roundtrip(self, csr_mat):
        children, aux = csr_mat.tree_flatten()
        restored = brainevent.CSR.tree_unflatten(aux, children)
        np.testing.assert_array_equal(restored.data, csr_mat.data)
        np.testing.assert_array_equal(restored.indices, csr_mat.indices)
        np.testing.assert_array_equal(restored.indptr, csr_mat.indptr)
        assert restored.shape == csr_mat.shape


# ===========================================================================
# 7. CSC buffer integration
# ===========================================================================

class TestCSCBuffers:
    @pytest.fixture
    def csc_mat(self):
        data = jnp.array([1.0, 2.0, 3.0, 4.0])
        indices = jnp.array([0, 1, 0, 1])
        indptr = jnp.array([0, 2, 3, 4])
        return brainevent.CSC((data, indices, indptr), shape=(2, 3))

    def test_csc_no_buffers_initially(self, csc_mat):
        assert csc_mat.buffers == {}
        assert not hasattr(csc_mat, 'diag_positions')

    def test_csc_pytree_roundtrip(self, csc_mat):
        csc_mat.register_buffer('diag_positions', jnp.array([0, 2]))
        children, aux = csc_mat.tree_flatten()
        restored = brainevent.CSC.tree_unflatten(aux, children)
        np.testing.assert_array_equal(restored.diag_positions, jnp.array([0, 2]))

    def test_csc_with_data_preserves_buffers(self, csc_mat):
        csc_mat.register_buffer('diag_positions', jnp.array([0, 2]))
        new_csc = csc_mat.with_data(jnp.array([10.0, 20.0, 30.0, 40.0]))
        np.testing.assert_array_equal(new_csc.diag_positions, jnp.array([0, 2]))

    def test_csc_apply_preserves_buffers(self, csc_mat):
        csc_mat.register_buffer('diag_positions', jnp.array([0, 2]))
        result = csc_mat.apply(lambda x: x * 3)
        np.testing.assert_array_equal(result.diag_positions, jnp.array([0, 2]))

    def test_csc_transpose_preserves_buffers(self, csc_mat):
        csc_mat.register_buffer('diag_positions', jnp.array([0, 2]))
        csr = csc_mat.transpose()
        assert 'diag_positions' in csr._buffer_registry
        np.testing.assert_array_equal(csr.diag_positions, jnp.array([0, 2]))


# ===========================================================================
# 8. COO buffer integration
# ===========================================================================

class TestCOOBuffers:
    @pytest.fixture
    def coo_mat(self):
        data = jnp.array([1.0, 2.0, 3.0])
        row = jnp.array([0, 1, 2])
        col = jnp.array([1, 0, 2])
        return brainevent.COO((data, row, col), shape=(3, 3))

    def test_coo_empty_buffers_by_default(self, coo_mat):
        assert coo_mat.buffers == {}

    def test_coo_register_and_roundtrip(self, coo_mat):
        coo_mat.register_buffer('my_cache', jnp.array([10, 20, 30]))
        children, aux = coo_mat.tree_flatten()
        restored = brainevent.COO.tree_unflatten(aux, children)
        assert 'my_cache' in restored._buffer_registry
        np.testing.assert_array_equal(restored.my_cache, jnp.array([10, 20, 30]))

    def test_coo_jit_preserves_buffers(self, coo_mat):
        coo_mat.register_buffer('tag', jnp.array(42))

        @jax.jit
        def f(m):
            return m

        result = f(coo_mat)
        assert int(result.tag) == 42


# ===========================================================================
# 9. diag_add integration (buffers carry through real operations)
# ===========================================================================

class TestDiagAddBufferIntegration:
    def test_diag_add_lazily_registers_diag_positions(self):
        n = 4
        data = jnp.ones(n * 2, dtype=jnp.float32)
        indices = jnp.array([0, 1, 1, 2, 2, 3, 3, 0], dtype=jnp.int32)
        indptr = jnp.array([0, 2, 4, 6, 8], dtype=jnp.int32)
        csr = brainevent.CSR((data, indices, indptr), shape=(n, n))

        assert not hasattr(csr, 'diag_positions')
        result = csr.diag_add(jnp.ones(n))
        assert hasattr(csr, 'diag_positions')
        assert 'diag_positions' in csr._buffer_registry
        assert 'diag_positions' in result._buffer_registry
        assert result.diag_positions is not None

    def test_diag_add_reuses_cached_positions(self):
        n = 3
        col_indices = [j for i in range(n) for j in range(n)]
        data = jnp.ones(n * n, dtype=jnp.float32)
        indices = jnp.array(col_indices, dtype=jnp.int32)
        indptr = jnp.array([i * n for i in range(n + 1)], dtype=jnp.int32)
        csr = brainevent.CSR((data, indices, indptr), shape=(n, n))

        result1 = csr.diag_add(jnp.ones(n))
        diag_pos1 = result1.diag_positions
        result2 = result1.diag_add(jnp.ones(n) * 2)
        diag_pos2 = result2.diag_positions
        np.testing.assert_array_equal(diag_pos1, diag_pos2)

    def test_diag_add_positions_survive_jit(self):
        n = 3
        col_indices = [j for i in range(n) for j in range(n)]
        data = jnp.ones(n * n, dtype=jnp.float32)
        indices = jnp.array(col_indices, dtype=jnp.int32)
        indptr = jnp.array([i * n for i in range(n + 1)], dtype=jnp.int32)
        csr = brainevent.CSR((data, indices, indptr), shape=(n, n))

        result = csr.diag_add(jnp.ones(n))

        @jax.jit
        def identity(m):
            return m

        jitted = identity(result)
        assert 'diag_positions' in jitted._buffer_registry
        np.testing.assert_array_equal(jitted.diag_positions, result.diag_positions)


# ===========================================================================
# 10. Multiple buffers on the same instance
# ===========================================================================

class TestMultipleBuffers:
    def test_three_buffers_roundtrip(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        obj.register_buffer('extra', jnp.array(99.0))
        obj.set_buffer('cached_sum', jnp.array(3.14))
        obj.set_buffer('label', 'hello')

        children, aux = obj.tree_flatten()
        restored = _SimpleBuffered.tree_unflatten(aux, children)

        assert float(restored.cached_sum) == pytest.approx(3.14)
        assert restored.label == 'hello'
        assert float(restored.extra) == 99.0
        assert restored._buffer_registry == {'cached_sum', 'label', 'extra'}

    def test_independent_instances_have_separate_registries(self):
        a = _SimpleBuffered(1.0, shape=(2, 2))
        b = _SimpleBuffered(2.0, shape=(3, 3))
        a.register_buffer('only_on_a', 10)
        assert 'only_on_a' in a._buffer_registry
        assert 'only_on_a' not in b._buffer_registry

    def test_csr_multiple_custom_buffers(self):
        data = jnp.array([1.0, 2.0])
        indices = jnp.array([0, 1])
        indptr = jnp.array([0, 1, 2])
        csr = brainevent.CSR((data, indices, indptr), shape=(2, 2))
        csr.register_buffer('my_mask', jnp.array([True, False]))
        csr.register_buffer('scale_factor', jnp.array(0.5))

        children, aux = csr.tree_flatten()
        restored = brainevent.CSR.tree_unflatten(aux, children)
        np.testing.assert_array_equal(restored.my_mask, jnp.array([True, False]))
        assert float(restored.scale_factor) == 0.5
        assert 'my_mask' in restored._buffer_registry
        assert 'scale_factor' in restored._buffer_registry


# ===========================================================================
# 11. Edge cases
# ===========================================================================

class TestEdgeCases:
    def test_buffer_value_array_types(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        for val in [jnp.array([1, 2, 3]), np.array([4.0]), 42, 'text', None]:
            obj.set_buffer('cached_sum', val)
            assert obj.buffers['cached_sum'] is val or np.array_equal(obj.buffers['cached_sum'], val)

    def test_buffer_with_jax_array_preserves_shape_dtype(self):
        arr = jnp.zeros((4, 5), dtype=jnp.float16)
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        obj.set_buffer('cached_sum', arr)
        children, aux = obj.tree_flatten()
        restored = _SimpleBuffered.tree_unflatten(aux, children)
        assert restored.cached_sum.shape == (4, 5)
        assert restored.cached_sum.dtype == jnp.float16

    def test_register_same_name_twice_is_idempotent_in_registry(self):
        obj = _SimpleBuffered(1.0, shape=(2, 2))
        obj.register_buffer('cached_sum', 1)
        obj.register_buffer('cached_sum', 2)
        assert obj.cached_sum == 2
        assert sum(1 for n in obj._buffer_registry if n == 'cached_sum') == 1

    def test_buffers_kwarg_with_unknown_names_registers_them(self):
        bufs = {'custom_a': 1, 'custom_b': jnp.array([2.0])}
        obj = _SimpleBuffered(1.0, shape=(2, 2), buffers=bufs)
        assert obj.custom_a == 1
        np.testing.assert_array_equal(obj.custom_b, jnp.array([2.0]))
        assert 'custom_a' in obj._buffer_registry
        assert 'custom_b' in obj._buffer_registry



# Copyright 2024 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import brainstate
import jax
import jax.numpy as jnp
import numpy as np
import pytest

# Keep GPU matmul reference numerics stable (avoid TF32 drift in dense @ B checks).
if jax.default_backend() == 'gpu' and jax.config.jax_default_matmul_precision is None:
    jax.config.update('jax_default_matmul_precision', 'highest')

import brainevent
from brainevent._test_util import allclose, gen_events
from brainevent._typing import MatrixShape

platform = jax.default_backend()

if platform == 'cpu':
    shapes = [
        (200, 300),
        (100, 500),
    ]
else:
    shapes = [
        (2000, 3000),
        (1000, 5000),
    ]


class Test_JITC_RC_Conversion:
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    def test_jitvec(self, shape: MatrixShape, corder):
        jitcr = brainevent.JITCUniformR((-0.1, 0.1, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        vector = jnp.asarray(np.random.rand(shape[1]))

        out1 = jitcr @ vector
        out2 = vector @ jitcc
        assert allclose(out1, out2)
        jax.block_until_ready((vector, out1, out2))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    def test_vecjit(self, shape: MatrixShape, corder):
        jitcr = brainevent.JITCUniformR((-0.1, 0.1, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        vector = jnp.asarray(np.random.rand(shape[0]))

        out1 = vector @ jitcr
        out2 = jitcc @ vector
        assert allclose(out1, out2)
        jax.block_until_ready((vector, out1, out2))

    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    def test_jitmat(self, k, shape: MatrixShape, corder):
        jitcr = brainevent.JITCUniformR((-0.1, 0.1, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        matrix = jnp.asarray(np.random.rand(shape[1], k))

        out1 = jitcr @ matrix
        out2 = (matrix.T @ jitcc).T
        assert allclose(out1, out2, atol=1e-3, rtol=1e-3)
        jax.block_until_ready((matrix, out1, out2))

    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    def test_matjit(self, k, shape: MatrixShape, corder):
        jitcr = brainevent.JITCUniformR((-0.1, 0.1, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        matrix = jnp.asarray(np.random.rand(k, shape[0]))

        out1 = matrix @ jitcr
        out2 = (jitcc @ matrix.T).T
        assert allclose(out1, out2, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((matrix, out1, out2))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    def test_jitvec_event(self, shape: MatrixShape, corder):
        jitcr = brainevent.JITCUniformR((-0.1, 0.1, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        vector = gen_events(shape[1])

        out1 = jitcr @ vector
        out2 = vector @ jitcc
        assert allclose(out1, out2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((out1, out2))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    def test_vecjit_event(self, shape: MatrixShape, corder):
        jitcr = brainevent.JITCUniformR((-0.1, 0.1, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        vector = gen_events(shape[0])

        out1 = vector @ jitcr
        out2 = jitcc @ vector
        assert allclose(out1, out2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((out1, out2))

    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('asbool', [True, False])
    def test_jitmat_event(self, k, shape: MatrixShape, corder, asbool):
        jitcr = brainevent.JITCUniformR((-0.1, 0.1, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        matrix = gen_events([shape[1], k], asbool=asbool)

        out1 = jitcr @ matrix
        out2 = (matrix.value.T @ jitcc).T
        assert allclose(out1, out2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((out1, out2))

    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('asbool', [True, False])
    def test_matjit_event(self, k, shape: MatrixShape, corder, asbool):
        jitcr = brainevent.JITCUniformR((-0.1, 0.1, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        matrix = gen_events([k, shape[0]], asbool=asbool)

        out1 = matrix @ jitcr
        out2 = (jitcc @ matrix.value.T).T
        assert allclose(out1, out2, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((out1, out2))


class Test_JITC_Operator_Behavior:
    @pytest.mark.parametrize('corder', [True, False])
    def test_jitc_uniform_r_operator_behavior(self, corder):
        shape = (20, 30)
        mat = brainevent.JITCUniformR((-1.5, 1.5, 0.1, 123), shape=shape, corder=corder)
        dense = mat.todense()

        left_vec = gen_events(shape[0], asbool=False).value
        right_vec = gen_events(shape[1], asbool=False).value
        left_mat = gen_events((5, shape[0]), asbool=False).value
        right_mat = gen_events((shape[1], 4), asbool=False).value

        r1 = left_vec @ mat
        r2 = left_vec @ dense
        r3 = mat @ right_vec
        r4 = dense @ right_vec
        r5 = left_mat @ mat
        r6 = left_mat @ dense
        r7 = mat @ right_mat
        r8 = dense @ right_mat
        assert allclose(r1, r2, atol=1e-4, rtol=1e-4)
        assert allclose(r3, r4, atol=1e-4, rtol=1e-4)
        assert allclose(r5, r6, atol=1e-4, rtol=1e-4)
        assert allclose(r7, r8, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((dense, left_vec, right_vec, left_mat, right_mat, r1, r2, r3, r4, r5, r6, r7, r8))

    @pytest.mark.parametrize('corder', [True, False])
    def test_jitc_uniform_c_operator_behavior(self, corder):
        shape = (20, 30)
        mat = brainevent.JITCUniformC((-1.5, 1.5, 0.1, 123), shape=shape, corder=corder)
        dense = mat.todense()

        left_vec = gen_events(shape[0], asbool=False).value
        right_vec = gen_events(shape[1], asbool=False).value
        left_mat = gen_events((5, shape[0]), asbool=False).value
        right_mat = gen_events((shape[1], 4), asbool=False).value

        r1 = left_vec @ mat
        r2 = left_vec @ dense
        r3 = mat @ right_vec
        r4 = dense @ right_vec
        r5 = left_mat @ mat
        r6 = left_mat @ dense
        r7 = mat @ right_mat
        r8 = dense @ right_mat
        assert allclose(r1, r2, atol=1e-4, rtol=1e-4)
        assert allclose(r3, r4, atol=1e-4, rtol=1e-4)
        assert allclose(r5, r6, atol=1e-4, rtol=1e-4)
        assert allclose(r7, r8, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((dense, left_vec, right_vec, left_mat, right_mat, r1, r2, r3, r4, r5, r6, r7, r8))

    @pytest.mark.parametrize('corder', [True, False])
    def test_jitc_uniform_r_transpose_operator_behavior(self, corder):
        shape = (20, 30)
        mat = brainevent.JITCUniformR((-1.5, 1.5, 0.1, 123), shape=shape, corder=corder).T
        dense = mat.todense()

        left_vec = jnp.asarray(np.random.rand(shape[1]))
        right_vec = jnp.asarray(np.random.rand(shape[0]))
        left_mat = jnp.asarray(np.random.rand(5, shape[1]))
        right_mat = jnp.asarray(np.random.rand(shape[0], 4))

        r1 = left_vec @ mat
        r2 = left_vec @ dense
        r3 = mat @ right_vec
        r4 = dense @ right_vec
        r5 = left_mat @ mat
        r6 = left_mat @ dense
        r7 = mat @ right_mat
        r8 = dense @ right_mat
        assert allclose(r1, r2, atol=1e-4, rtol=1e-4)
        assert allclose(r3, r4, atol=1e-4, rtol=1e-4)
        assert allclose(r5, r6, atol=1e-4, rtol=1e-4)
        assert allclose(r7, r8, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((dense, left_vec, right_vec, left_mat, right_mat, r1, r2, r3, r4, r5, r6, r7, r8))

    @pytest.mark.parametrize('corder', [True, False])
    def test_jitc_uniform_c_transpose_operator_behavior(self, corder):
        shape = (20, 30)
        mat = brainevent.JITCUniformC((-1.5, 1.5, 0.1, 123), shape=shape, corder=corder).T
        dense = mat.todense()

        left_vec = jnp.asarray(np.random.rand(shape[1]))
        right_vec = jnp.asarray(np.random.rand(shape[0]))
        left_mat = jnp.asarray(np.random.rand(5, shape[1]))
        right_mat = jnp.asarray(np.random.rand(shape[0], 4))

        r1 = left_vec @ mat
        r2 = left_vec @ dense
        r3 = mat @ right_vec
        r4 = dense @ right_vec
        r5 = left_mat @ mat
        r6 = left_mat @ dense
        r7 = mat @ right_mat
        r8 = dense @ right_mat
        assert allclose(r1, r2, atol=1e-4, rtol=1e-4)
        assert allclose(r3, r4, atol=1e-4, rtol=1e-4)
        assert allclose(r5, r6, atol=1e-4, rtol=1e-4)
        assert allclose(r7, r8, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((dense, left_vec, right_vec, left_mat, right_mat, r1, r2, r3, r4, r5, r6, r7, r8))

    @pytest.mark.parametrize('cls', [brainevent.JITCUniformR, brainevent.JITCUniformC])
    def test_jitc_uniform_unit_operator_behavior(self, cls):
        import brainunit as u

        shape = (20, 30)
        weight = 2.1 * u.mV
        mat = cls((-weight, weight, 0.2, 123), shape=shape)
        dense = mat.todense()

        right_vec = jnp.asarray(np.random.rand(shape[1]))
        left_vec = jnp.asarray(np.random.rand(shape[0]))

        r1 = mat @ right_vec
        r2 = dense @ right_vec
        r3 = left_vec @ mat
        r4 = left_vec @ dense
        assert u.math.allclose(
            r1,
            r2,
            rtol=1e-4 * u.get_unit(r2),
            atol=1e-4 * u.get_unit(r2),
        )
        assert u.math.allclose(
            r3,
            r4,
            rtol=1e-4 * u.get_unit(r4),
            atol=1e-4 * u.get_unit(r4),
        )
        jax.block_until_ready((right_vec, left_vec))


class Test_JITC_To_Dense:
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('transpose', [True, False])
    @pytest.mark.parametrize('corder', [True, False])
    def test_todense(self, shape: MatrixShape, transpose, corder):
        jitcr = brainevent.JITCUniformR((-0.1, 0.1, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        out1 = jitcr.todense()
        out2 = jitcc.todense().T
        out3 = jitcr.T.todense().T
        out4 = jitcc.T.todense()
        assert allclose(out1, out2)
        assert allclose(out1, out3)
        assert allclose(out1, out4)
        jax.block_until_ready((out1, out2, out3, out4))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('wlow', [-1., 0.])
    @pytest.mark.parametrize('whigh', [1., 2.])
    def test_vjp(self, shape, corder, wlow, whigh):
        base = brainevent.JITCUniformR((0.0, 1.0, 0.1, 123), shape=shape, corder=corder).todense()

        def f_dense_vjp(wlow, whigh):
            res = base * (whigh - wlow) + wlow
            return res

        ct = brainstate.random.random(shape)
        primals, f_vjp = jax.vjp(f_dense_vjp, wlow, whigh)
        true_wlow_grad, true_whigh_grad = f_vjp(ct)

        expected_wlow_grad = (ct * (-base + 1.)).sum()
        expected_whigh_grad = (ct * base).sum()

        assert allclose(true_wlow_grad, expected_wlow_grad)
        assert allclose(true_whigh_grad, expected_whigh_grad)

        print(true_wlow_grad, true_whigh_grad)
        print(expected_wlow_grad, expected_whigh_grad)

        def f_jitc_vjp(wlow, whigh):
            mat = brainevent.JITCUniformR((wlow, whigh, 0.1, 123), shape=shape, corder=corder)
            return mat.todense()

        primals, f_vjp2 = jax.vjp(f_jitc_vjp, wlow, whigh)
        jitc_wlow_grad, jitc_whigh_grad = f_vjp2(ct)

        assert allclose(true_wlow_grad, jitc_wlow_grad)
        assert allclose(true_whigh_grad, jitc_whigh_grad)
        jax.block_until_ready(
            (base, ct, primals, true_wlow_grad, true_whigh_grad, expected_wlow_grad, expected_whigh_grad,
             jitc_wlow_grad, jitc_whigh_grad))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('wlow', [-1., 0.])
    @pytest.mark.parametrize('whigh', [1., 2.])
    def test_jvp(self, shape, corder, wlow, whigh):
        base = brainevent.JITCUniformR((0.0, 1.0, 0.1, 123), shape=shape, corder=corder).todense()
        tagents = (brainstate.random.random(), brainstate.random.random())

        def f_dense_jvp(wlow, whigh):
            res = base * (whigh - wlow) + wlow
            return res

        def f_jitc_jvp(wlow, whigh):
            mat = brainevent.JITCUniformR((wlow, whigh, 0.1, 123), shape=shape, corder=corder)
            return mat.todense()

        primals1, true_grad = jax.jvp(f_dense_jvp, (wlow, whigh), tagents)
        primals2, jitc_grad = jax.jvp(f_jitc_jvp, (wlow, whigh), tagents)
        assert allclose(true_grad, jitc_grad)
        jax.block_until_ready((base, tagents[0], tagents[1], primals1, true_grad, primals2, jitc_grad))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('wlow', [-1., 0.])
    @pytest.mark.parametrize('whigh', [1., 2.])
    @pytest.mark.parametrize('dwlow', [1., 2.])
    def test_jvp_wlow(self, shape, corder, wlow, whigh, dwlow):
        base = brainevent.JITCUniformR((0.0, 1.0, 0.1, 123), shape=shape, corder=corder).todense()

        def f_dense_jvp(wlow):
            res = base * (whigh - wlow) + wlow
            return res

        primals1, true_grad = jax.jvp(f_dense_jvp, (wlow,), (dwlow,))
        expected_grad = (-base + 1.) * dwlow
        assert allclose(true_grad, expected_grad)

        def f_jitc_jvp(wlow):
            mat = brainevent.JITCUniformR((wlow, whigh, 0.1, 123), shape=shape, corder=corder)
            return mat.todense()

        primals2, jitc_grad = jax.jvp(f_jitc_jvp, (wlow,), (dwlow,))
        assert allclose(true_grad, jitc_grad)
        jax.block_until_ready((base, primals1, true_grad, expected_grad, primals2, jitc_grad))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('wlow', [-1., 0.])
    @pytest.mark.parametrize('whigh', [1., 2.])
    @pytest.mark.parametrize('dw_high', [1., 2.])
    def test_jvp_whigh(self, shape, corder, wlow, whigh, dw_high):
        base = brainevent.JITCUniformR((0.0, 1.0, 0.1, 123), shape=shape, corder=corder).todense()

        def f_dense_jvp(whigh):
            res = base * (whigh - wlow) + wlow
            return res

        primals1, true_grad = jax.jvp(f_dense_jvp, (whigh,), (dw_high,))
        expected_grad = base * dw_high
        assert allclose(true_grad, expected_grad)

        def f_jitc_jvp(whigh):
            mat = brainevent.JITCUniformR((wlow, whigh, 0.1, 123), shape=shape, corder=corder)
            return mat.todense()

        primals2, jitc_grad = jax.jvp(f_jitc_jvp, (whigh,), (dw_high,))
        assert allclose(true_grad, jitc_grad)
        jax.block_until_ready((base, primals1, true_grad, expected_grad, primals2, jitc_grad))


class Test_JITC_Uniform_Validation:
    @pytest.mark.parametrize('cls', [brainevent.JITCUniformR, brainevent.JITCUniformC])
    @pytest.mark.parametrize('prob', [-0.1, 1.1, float('nan')])
    def test_invalid_prob_raises(self, cls, prob):
        with pytest.raises(ValueError, match='prob'):
            cls((-1.0, 1.0, prob, 123), shape=(8, 6))

    @pytest.mark.parametrize('cls', [brainevent.JITCUniformR, brainevent.JITCUniformC])
    def test_invalid_bounds_raises(self, cls):
        with pytest.raises(ValueError, match='wlow'):
            cls((1.0, -1.0, 0.1, 123), shape=(8, 6))

    # @pytest.mark.parametrize('cls', [brainevent.JITCUniformR, brainevent.JITCUniformC])
    # @pytest.mark.parametrize('corder', [True, False])
    # def test_zero_prob_dense_matvec_matmat(self, cls, corder):
    #     shape = (8, 6)
    #     mat = cls((-1.0, 1.0, 0.0, 123), shape=shape, corder=corder)
    #
    #     dense = mat.todense()
    #     assert allclose(dense, jnp.zeros_like(dense))
    #
    #     vec = jnp.ones(shape[1])
    #     out_mv = mat @ vec
    #     assert allclose(out_mv, jnp.zeros_like(out_mv))
    #
    #     B = jnp.ones((shape[1], 4))
    #     out_mm = mat @ B
    #     assert allclose(out_mm, jnp.zeros_like(out_mm))
    #     jax.block_until_ready((dense, vec, out_mv, B, out_mm))



# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# -*- coding: utf-8 -*-

import os

os.environ['JAX_TRACEBACK_FILTERING'] = 'off'

import brainstate
import jax
import jax.numpy as jnp

from brainevent._csr.diag_add import (
    csr_diag_position,
    csr_diag_position_v2,
    csr_diag_add,
    csr_diag_add_v2,
    csr_diag_add_call,
)
from brainevent._csr.test_util import get_csr


class TestCSRDiagPosition:
    def test_basic_position(self):
        """Test basic diagonal position finding."""
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        positions = csr_diag_position(indptr, indices, (m, n))

        assert positions.shape == (min(m, n),)
        assert jnp.issubdtype(positions.dtype, jnp.integer)

        # Verify positions: for each diagonal element, check if the position is correct
        for i in range(min(m, n)):
            pos = positions[i]
            if pos >= 0:
                # Check that indices[pos] == i (the column index should match the diagonal)
                assert indices[pos] == i

    def test_rectangular_matrix(self):
        """Test diagonal position finding for rectangular matrices."""
        for m, n in [(10, 5), (5, 10)]:
            indptr, indices = get_csr(m, n, 0.5)
            positions = csr_diag_position(indptr, indices, (m, n))

            assert positions.shape == (min(m, n),)

    def test_position_v2(self):
        """Test the v2 version of diagonal position finding."""
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        csr_pos, diag_pos = csr_diag_position_v2(indptr, indices, shape=(m, n))

        assert csr_pos.ndim == 1
        if diag_pos is not None:
            assert diag_pos.ndim == 1
            assert len(csr_pos) == len(diag_pos)

    def test_no_diagonal_elements(self):
        """Test case where there are no diagonal elements."""
        # Create CSR with off-diagonal elements only
        indptr = jnp.array([0, 1, 2, 3])
        indices = jnp.array([1, 2, 0])  # No diagonal elements

        positions = csr_diag_position(indptr, indices, (3, 3))

        # All positions should be -1
        assert jnp.all(positions == -1)


class TestCSRDiagAdd:
    def test_basic_add(self):
        """Test basic diagonal addition."""
        m, n = 10, 10
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(jnp.float32)
        diag_position = csr_diag_position(indptr, indices, (m, n))
        diag_value = brainstate.random.rand(m).astype(jnp.float32)

        result = csr_diag_add(csr_value, diag_position, diag_value)

        assert result.shape == csr_value.shape
        assert result.dtype == csr_value.dtype

        # Verify the diagonal values were added correctly
        for i in range(m):
            pos = diag_position[i]
            if pos >= 0:
                expected = csr_value[pos] + diag_value[i]
                assert jnp.allclose(result[pos], expected)

    def test_add_with_negative_positions(self):
        """Test diagonal addition ignores negative positions."""
        csr_value = jnp.array([1.0, 2.0, 3.0, 4.0])
        diag_position = jnp.array([0, -1, 2])  # -1 should be ignored
        diag_value = jnp.array([10.0, 20.0, 30.0])

        result = csr_diag_add(csr_value, diag_position, diag_value)

        expected = jnp.array([11.0, 2.0, 33.0, 4.0])
        assert jnp.allclose(result, expected)

    def test_add_v2(self):
        """Test the v2 version of diagonal addition."""
        m, n = 10, 10
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(jnp.float32)
        positions = csr_diag_position_v2(indptr, indices, shape=(m, n))
        diag_value = brainstate.random.rand(m).astype(jnp.float32)

        result = csr_diag_add_v2(csr_value, positions, diag_value)

        assert result.shape == csr_value.shape
        assert result.dtype == csr_value.dtype

    def test_dtype_consistency(self):
        """Test that different dtypes work correctly."""
        # Only test float32 since float64 may not be available in all JAX configurations
        dtype = jnp.float32
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(dtype)
        diag_position = csr_diag_position(indptr, indices, (m, n))
        diag_value = brainstate.random.rand(m).astype(dtype)

        result = csr_diag_add(csr_value, diag_position, diag_value)

        assert result.dtype == dtype


class TestCSRDiagAddGradients:
    def test_jvp_csr_value(self):
        """Test JVP with respect to csr_value."""
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(jnp.float32)
        diag_position = csr_diag_position(indptr, indices, (m, n))
        diag_value = brainstate.random.rand(m).astype(jnp.float32)

        def f(x):
            return csr_diag_add(x, diag_position, diag_value)

        primals, tangents = jax.jvp(f, (csr_value,), (jnp.ones_like(csr_value),))

        assert primals.shape == csr_value.shape
        assert tangents.shape == csr_value.shape

    def test_jvp_csr_value_is_identity(self):
        """Test that JVP w.r.t. csr_value is the identity (tangent == perturbation)."""
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(jnp.float32)
        diag_position = csr_diag_position(indptr, indices, (m, n))
        diag_value = brainstate.random.rand(m).astype(jnp.float32)

        perturbation = brainstate.random.rand(nse).astype(jnp.float32)

        def f(x):
            return csr_diag_add(x, diag_position, diag_value)

        _, tangents = jax.jvp(f, (csr_value,), (perturbation,))
        assert jnp.allclose(tangents, perturbation), (
            "JVP w.r.t. csr_value should be identity: tangent must equal perturbation"
        )

    def test_jvp_diag_value(self):
        """Test JVP with respect to diag_value."""
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(jnp.float32)
        diag_position = csr_diag_position(indptr, indices, (m, n))
        diag_value = brainstate.random.rand(m).astype(jnp.float32)

        def f(x):
            return csr_diag_add(csr_value, diag_position, x)

        primals, tangents = jax.jvp(f, (diag_value,), (jnp.ones_like(diag_value),))

        assert primals.shape == csr_value.shape
        assert tangents.shape == csr_value.shape

    def test_jvp_diag_value_scatter(self):
        """Test that JVP w.r.t. diag_value scatters the perturbation correctly."""
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(jnp.float32)
        diag_position = csr_diag_position(indptr, indices, (m, n))
        diag_value = brainstate.random.rand(m).astype(jnp.float32)

        perturbation = brainstate.random.rand(m).astype(jnp.float32)

        def f(x):
            return csr_diag_add(csr_value, diag_position, x)

        _, tangents = jax.jvp(f, (diag_value,), (perturbation,))

        # Expected: zeros.at[diag_position].add(perturbation)
        expected = jnp.zeros_like(csr_value)
        for i in range(m):
            pos = int(diag_position[i])
            if pos >= 0:
                expected = expected.at[pos].add(perturbation[i])

        assert jnp.allclose(tangents, expected), (
            "JVP w.r.t. diag_value should scatter perturbation to diag positions"
        )

    def test_jvp_both(self):
        """Test JVP with respect to both csr_value and diag_value."""
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(jnp.float32)
        diag_position = csr_diag_position(indptr, indices, (m, n))
        diag_value = brainstate.random.rand(m).astype(jnp.float32)

        def f(csr, diag):
            return csr_diag_add(csr, diag_position, diag)

        primals, tangents = jax.jvp(
            f,
            (csr_value, diag_value),
            (jnp.ones_like(csr_value), jnp.ones_like(diag_value))
        )

        assert primals.shape == csr_value.shape
        assert tangents.shape == csr_value.shape

    def test_jvp_combined_vs_reference(self):
        """Test combined JVP against manual reference computation."""
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(jnp.float32)
        diag_position = csr_diag_position(indptr, indices, (m, n))
        diag_value = brainstate.random.rand(m).astype(jnp.float32)

        csr_dot = brainstate.random.rand(nse).astype(jnp.float32)
        diag_dot = brainstate.random.rand(m).astype(jnp.float32)

        def f(csr, diag):
            return csr_diag_add(csr, diag_position, diag)

        _, tangents = jax.jvp(f, (csr_value, diag_value), (csr_dot, diag_dot))

        # Reference: d/dt [csr + scatter(diag)] = csr_dot + scatter(diag_dot)
        expected = csr_dot
        for i in range(m):
            pos = int(diag_position[i])
            if pos >= 0:
                expected = expected.at[pos].add(diag_dot[i])

        assert jnp.allclose(tangents, expected, atol=1e-6), (
            "Combined JVP should equal csr_dot + scatter(diag_dot)"
        )


class TestCSRDiagAddJIT:
    def test_jit_basic(self):
        """Test JIT compilation of csr_diag_add."""
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(jnp.float32)
        diag_position = csr_diag_position(indptr, indices, (m, n))
        diag_value = brainstate.random.rand(m).astype(jnp.float32)

        @jax.jit
        def f(csr, diag):
            return csr_diag_add(csr, diag_position, diag)

        result = f(csr_value, diag_value)

        # Compare with non-JIT version
        expected = csr_diag_add(csr_value, diag_position, diag_value)

        assert jnp.allclose(result, expected)

    def test_jit_call_function(self):
        """Test JIT compilation of csr_diag_add_call."""
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(jnp.float32)
        diag_position = csr_diag_position(indptr, indices, (m, n))
        diag_value = brainstate.random.rand(m).astype(jnp.float32)

        @jax.jit
        def f(csr, pos, diag):
            return csr_diag_add_call(csr, pos, diag)

        result = f(csr_value, diag_position, diag_value)

        assert result[0].shape == csr_value.shape

    def test_jit_with_grad(self):
        """Test JIT compilation with gradients."""
        m, n = 5, 5
        indptr, indices = get_csr(m, n, 0.5)
        nse = indices.shape[0]

        csr_value = brainstate.random.rand(nse).astype(jnp.float32)
        diag_position = csr_diag_position(indptr, indices, (m, n))
        diag_value = brainstate.random.rand(m).astype(jnp.float32)

        @jax.jit
        def f(csr, diag):
            return jnp.sum(csr_diag_add(csr, diag_position, diag))

        # JVP should work under JIT
        _, tangent = jax.jvp(f, (csr_value, diag_value), (jnp.ones_like(csr_value), jnp.ones_like(diag_value)))

        assert tangent.shape == ()



# Copyright 2026 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import os

os.environ['JAX_TRACEBACK_FILTERING'] = 'off'

import brainstate
import braintools
import jax
import jax.numpy as jnp
import pytest

from brainevent._csr.yw2y import csrmv_yw2y, csrmv_yw2y_p
from brainevent._csr.test_util import get_csr

platform = jax.default_backend()
CSRMV_YW2Y_IMPLEMENTATIONS = tuple(csrmv_yw2y_p.available_backends(platform))


def _row_ids_from_indptr(indptr):
    indptr = jnp.asarray(indptr)
    counts = jnp.diff(indptr)
    return jnp.repeat(jnp.arange(counts.shape[0], dtype=indptr.dtype), counts)


@pytest.mark.skipif(
    not CSRMV_YW2Y_IMPLEMENTATIONS,
    reason=f'No csrmv_yw2y implementation on platform={platform}',
)
class TestCSRMVYw2y:
    @pytest.mark.parametrize('implementation', CSRMV_YW2Y_IMPLEMENTATIONS)
    @pytest.mark.parametrize('shape', [(100, 200), (200, 400)])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr(self, implementation, shape, transpose):
        m, n = shape
        indptr, indices = get_csr(m, n, 0.5)

        data = braintools.init.Normal(0.0, 1.0)(indices.shape)
        y = brainstate.random.rand(n) if transpose else brainstate.random.rand(m)

        result = csrmv_yw2y(y, data, indices, indptr, shape=(m, n), transpose=transpose, backend=implementation)

        if transpose:
            expected = data * y[indices]
        else:
            row_ids = _row_ids_from_indptr(indptr)
            expected = data * y[row_ids]

        assert jnp.allclose(result, expected, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((data, y, indptr, indices, result, expected))



# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# -*- coding: utf-8 -*-


from typing import Sequence, Optional

import brainunit as u
import jax
import jax.numpy as jnp
import numpy as np
from jax.interpreters import ad

from brainevent._misc import generate_block_dim, namescope
from brainevent._op import jaxinfo_to_warpinfo, numba_kernel, XLACustomKernel, general_batching_rule
from brainevent._op.benchmark import BenchmarkConfig
from brainevent._sddmm import sddmm_coo_indices
from brainevent._typing import Data, Row, Col, MatrixShape
from .float import coomv, coomm

__all__ = [
    'binary_coomv',
    'binary_coomv_p',
    'binary_coomm',
    'binary_coomm_p',
]


@namescope(static_argnames=("shape", "transpose"))
def binary_coomv(
    data: Data,
    row: Row,
    col: Col,
    v: Data,
    *,
    shape: MatrixShape,
    transpose: bool = False,
    backend: Optional[str] = None,
) -> Data:
    """
    Perform event-driven COO sparse matrix-vector multiplication.

    Computes the product of a sparse matrix stored in COO (Coordinate) format
    and a dense vector, where the vector ``v`` is treated as containing binary
    events (spikes). Only entries corresponding to active (nonzero / ``True``)
    events contribute to the output, making this operation efficient for
    spike-based neural network simulations.

    Mathematically, with ``transpose=False`` the operation is

    ``result[i] = sum_j A[i, j] * (v[j] > 0)``

    and with ``transpose=True``:

    ``result[j] = sum_i A[i, j] * (v[i] > 0)``

    where ``A`` is the sparse matrix defined by (``data``, ``row``, ``col``).

    Parameters
    ----------
    data : jax.Array
        The non-zero weight values of the sparse matrix.  Can be either a
        scalar (shape ``(1,)``, homogeneous weights) or a 1-D array of
        length ``nnz`` (heterogeneous weights).
    row : jax.Array
        1-D array of row indices for each non-zero element, with length
        ``nnz``.
    col : jax.Array
        1-D array of column indices for each non-zero element, with length
        ``nnz``.
    v : jax.Array
        Dense input vector treated as binary events. Elements are
        interpreted as active when ``True`` (boolean dtype) or ``> 0``
        (numeric dtype).  Shape must be ``(shape[0],)`` when
        ``transpose=True`` or ``(shape[1],)`` when ``transpose=False``.
    shape : tuple of int
        The ``(m, k)`` shape of the sparse matrix.
    transpose : bool, optional
        If ``True``, multiply by the transpose of the sparse matrix,
        i.e. ``A^T @ v``.  Default is ``False``.
    backend : str or None, optional
        Compute backend to use (e.g. ``'numba'``, ``'warp'``,
        ``'pallas'``).  When ``None`` the backend is chosen automatically.

    Returns
    -------
    jax.Array
        Result vector of the matrix-vector product.  Shape is
        ``(shape[0],)`` when ``transpose=False`` or ``(shape[1],)`` when
        ``transpose=True``.

    Raises
    ------
    ValueError
        If ``row`` or ``col`` are not 1-D, have mismatched lengths, ``v``
        is not 1-D, ``data`` is not scalar or 1-D, or array dimensions
        are incompatible with ``shape`` and ``transpose``.

    See Also
    --------
    binary_coomm : Event-driven COO sparse matrix-matrix multiplication.
    coomv : Standard (non-event-driven) COO sparse matrix-vector multiplication.

    Notes
    -----
    Physical units attached via ``brainunit`` are automatically split off
    before the computation and re-applied to the result.

    This function supports automatic differentiation (JVP and transpose
    rules), ``vmap`` batching, and multiple hardware backends (CPU via
    Numba, GPU via Warp or Pallas/Triton, TPU via Pallas/Mosaic).

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._coo.binary import binary_coomv
        >>> data = jnp.array([0.5])          # homogeneous weight
        >>> row = jnp.array([0, 1, 2])
        >>> col = jnp.array([1, 0, 2])
        >>> v = jnp.array([True, False, True])  # binary events
        >>> binary_coomv(data, row, col, v, shape=(3, 3), transpose=False)
    """
    data, unitd = u.split_mantissa_unit(data)
    v, unitv = u.split_mantissa_unit(v)
    res = binary_coomv_p_call(
        data, row, col, v,
        shape=shape,
        transpose=transpose,
        backend=backend,
    )[0]
    return u.maybe_decimal(res * (unitd * unitv))


@namescope(static_argnames=("shape", "transpose"))
def binary_coomm(
    data: Data,
    row: Row,
    col: Col,
    B: Data,
    *,
    shape: MatrixShape,
    transpose: bool = False,
    backend: Optional[str] = None,
) -> Data:
    """
    Perform event-driven COO sparse matrix-matrix multiplication.

    Computes the product of a sparse matrix stored in COO (Coordinate) format
    and a dense matrix ``B``, where ``B`` is treated as containing binary
    events (spikes). Only entries of ``B`` that are active (nonzero /
    ``True``) contribute to the output.

    Mathematically, with ``transpose=False`` the operation is

    ``result[i, n] = sum_j A[i, j] * (B[j, n] > 0)``

    and with ``transpose=True``:

    ``result[j, n] = sum_i A[i, j] * (B[i, n] > 0)``

    where ``A`` is the sparse matrix defined by (``data``, ``row``, ``col``).

    Parameters
    ----------
    data : jax.Array
        The non-zero weight values of the sparse matrix.  Can be either a
        scalar (shape ``(1,)``, homogeneous weights) or a 1-D array of
        length ``nnz`` (heterogeneous weights).
    row : jax.Array
        1-D array of row indices for each non-zero element, with length
        ``nnz``.
    col : jax.Array
        1-D array of column indices for each non-zero element, with length
        ``nnz``.
    B : jax.Array
        Dense input matrix treated as binary events.  Shape must be
        ``(shape[0], n)`` when ``transpose=True`` or ``(shape[1], n)``
        when ``transpose=False``.
    shape : tuple of int
        The ``(m, k)`` shape of the sparse matrix.
    transpose : bool, optional
        If ``True``, multiply by the transpose of the sparse matrix,
        i.e. ``A^T @ B``.  Default is ``False``.
    backend : str or None, optional
        Compute backend to use (e.g. ``'numba'``, ``'warp'``,
        ``'pallas'``).  When ``None`` the backend is chosen automatically.

    Returns
    -------
    jax.Array
        Result matrix of the multiplication.  Shape is
        ``(shape[0], n)`` when ``transpose=False`` or ``(shape[1], n)``
        when ``transpose=True``.

    Raises
    ------
    ValueError
        If ``row`` or ``col`` are not 1-D, have mismatched lengths, ``B``
        is not 2-D, ``data`` is not scalar or 1-D, or array dimensions
        are incompatible with ``shape`` and ``transpose``.

    See Also
    --------
    binary_coomv : Event-driven COO sparse matrix-vector multiplication.
    coomm : Standard (non-event-driven) COO sparse matrix-matrix multiplication.

    Notes
    -----
    Physical units attached via ``brainunit`` are automatically split off
    before the computation and re-applied to the result.

    This function supports automatic differentiation (JVP and transpose
    rules), ``vmap`` batching, and multiple hardware backends (CPU via
    Numba, GPU via Warp or Pallas/Triton, TPU via Pallas/Mosaic).

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._coo.binary import binary_coomm
        >>> data = jnp.array([0.5])          # homogeneous weight
        >>> row = jnp.array([0, 1, 2])
        >>> col = jnp.array([1, 0, 2])
        >>> B = jnp.array([[True, False],
        ...                [False, True],
        ...                [True, True]])    # binary event matrix
        >>> binary_coomm(data, row, col, B, shape=(3, 3), transpose=False)
    """
    data, unitd = u.split_mantissa_unit(data)
    B, unitb = u.split_mantissa_unit(B)
    res = binary_coomm_p_call(
        data, row, col, B,
        shape=shape,
        transpose=transpose,
        backend=backend,
    )[0]
    return u.maybe_decimal(res * (unitd * unitb))


# =============================================================================
# COO Matrix-Vector Multiplication (coomv)
# =============================================================================


def _coomv_numba_kernel(
    weight_info: jax.ShapeDtypeStruct,
    vector_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    import numba

    if transpose:
        if weight_info.size == 1:
            # transpose=True, homogeneous
            if vector_info.dtype == jnp.bool_:
                # bool
                @numba.njit(fastmath=True)
                def mv(weights, row, col, v, posts):
                    posts[:] = 0.
                    w = weights[0]
                    for i in range(row.shape[0]):
                        if v[row[i]]:
                            posts[col[i]] += w
            else:
                @numba.njit(fastmath=True)
                def mv(weights, row, col, v, posts):
                    posts[:] = 0.
                    w = weights[0]
                    for i in range(row.shape[0]):
                        if v[row[i]] > 0.:
                            posts[col[i]] += w
        else:
            # transpose=True, heterogeneous
            if vector_info.dtype == jnp.bool_:
                # bool
                @numba.njit(fastmath=True)
                def mv(weights, row, col, v, posts):
                    posts[:] = 0.
                    for i in range(row.shape[0]):
                        if v[row[i]]:
                            posts[col[i]] += weights[i]
            else:
                @numba.njit(fastmath=True)
                def mv(weights, row, col, v, posts):
                    posts[:] = 0.
                    for i in range(row.shape[0]):
                        if v[row[i]] > 0.:
                            posts[col[i]] += weights[i]
    else:
        if weight_info.size == 1:
            # transpose=False, homogeneous
            if vector_info.dtype == jnp.bool_:
                # bool
                @numba.njit(fastmath=True)
                def mv(weights, row, col, v, posts):
                    posts[:] = 0.
                    w = weights[0]
                    for i in range(row.shape[0]):
                        if v[col[i]]:
                            posts[row[i]] += w
            else:
                @numba.njit(fastmath=True)
                def mv(weights, row, col, v, posts):
                    posts[:] = 0.
                    w = weights[0]
                    for i in range(row.shape[0]):
                        if v[col[i]] > 0.:
                            posts[row[i]] += w
        else:
            # transpose=False, heterogeneous
            if vector_info.dtype == jnp.bool_:
                # bool
                @numba.njit(fastmath=True)
                def mv(weights, row, col, v, posts):
                    posts[:] = 0.
                    for i in range(row.shape[0]):
                        if v[col[i]]:
                            posts[row[i]] += weights[i]
            else:
                @numba.njit(fastmath=True)
                def mv(weights, row, col, v, posts):
                    posts[:] = 0.
                    for i in range(row.shape[0]):
                        if v[col[i]] > 0.:
                            posts[row[i]] += weights[i]

    def kernel(weights, row, col, v):
        return numba_kernel(mv, outs=kwargs['outs'])(weights, row, col, v)

    return kernel


def _coomv_warp_kernel(
    weight_info: jax.ShapeDtypeStruct,
    vector_info: jax.ShapeDtypeStruct,
    row_info: jax.ShapeDtypeStruct,
    col_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    import warp
    from warp.jax_experimental import jax_kernel

    weight_warp_info = jaxinfo_to_warpinfo(weight_info)
    row_warp_info = jaxinfo_to_warpinfo(row_info)
    col_warp_info = jaxinfo_to_warpinfo(col_info)
    spike_warp_info = jaxinfo_to_warpinfo(vector_info)
    out_warp_info = jaxinfo_to_warpinfo(kwargs['outs'][0])

    if transpose:
        if weight_info.size == 1:
            # transpose=True, homogeneous
            if vector_info.dtype == jnp.bool_:
                # bool
                @warp.kernel
                def mv(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    v: spike_warp_info,
                    posts: out_warp_info
                ):
                    i = warp.tid()
                    w = weights[0]
                    if v[row[i]]:
                        warp.atomic_add(posts, col[i], w)
            else:
                @warp.kernel
                def mv(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    v: spike_warp_info,
                    posts: out_warp_info
                ):
                    i = warp.tid()
                    w = weights[0]
                    if v[row[i]] > 0.:
                        warp.atomic_add(posts, col[i], w)
        else:
            # transpose=True, heterogeneous
            if vector_info.dtype == jnp.bool_:
                # bool
                @warp.kernel
                def mv(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    v: spike_warp_info,
                    posts: out_warp_info
                ):
                    i = warp.tid()
                    if v[row[i]]:
                        warp.atomic_add(posts, col[i], weights[i])
            else:
                @warp.kernel
                def mv(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    v: spike_warp_info,
                    posts: out_warp_info
                ):
                    i = warp.tid()
                    if v[row[i]] > 0.:
                        warp.atomic_add(posts, col[i], weights[i])
    else:
        if weight_info.size == 1:
            # transpose=False, homogeneous
            if vector_info.dtype == jnp.bool_:
                # bool
                @warp.kernel
                def mv(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    v: spike_warp_info,
                    posts: out_warp_info
                ):
                    i = warp.tid()
                    w = weights[0]
                    if v[col[i]]:
                        warp.atomic_add(posts, row[i], w)
            else:
                @warp.kernel
                def mv(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    v: spike_warp_info,
                    posts: out_warp_info
                ):
                    i = warp.tid()
                    w = weights[0]
                    if v[col[i]] > 0.:
                        warp.atomic_add(posts, row[i], w)
        else:
            # transpose=False, heterogeneous
            if vector_info.dtype == jnp.bool_:
                # bool
                @warp.kernel
                def mv(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    v: spike_warp_info,
                    posts: out_warp_info
                ):
                    i = warp.tid()
                    if v[col[i]]:
                        warp.atomic_add(posts, row[i], weights[i])
            else:
                @warp.kernel
                def mv(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    v: spike_warp_info,
                    posts: out_warp_info
                ):
                    i = warp.tid()
                    if v[col[i]] > 0.:
                        warp.atomic_add(posts, row[i], weights[i])

    def kernel(weights, row, col, v):
        dim = row_info.shape[0]
        out_info = kwargs['outs'][0]
        fn = jax_kernel(mv, launch_dims=[dim], num_outputs=1, in_out_argnames=['posts'])
        return fn(weights, row, col, v, jnp.zeros(out_info.shape, out_info.dtype))

    return kernel


def _coomv_pallas_gpu_kernel(
    weight_info: jax.ShapeDtypeStruct,
    row_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    from jax.experimental import pallas as pl
    from jax.experimental.pallas import triton as plgpu
    from jax.experimental.pallas.triton import atomic_add

    nnz = row_info.shape[0]
    block_dim = generate_block_dim(nnz)
    block_dim = 32 if block_dim < 32 else block_dim

    if transpose:
        if weight_info.size == 1:
            # coo.T @ v (homogeneous weights)
            #
            # coo: [m, k]
            # v: [m]
            # result: [k]
            #
            def mv(
                data_ref,  # [1]
                row_ref,  # [nnz]
                col_ref,  # [nnz]
                vector_ref,  # [m]
                _,  # [k]
                posts_ref,  # [k]
            ):
                i = pl.program_id(0)
                i_start = i * block_dim
                mask = i_start + jnp.arange(block_dim) < nnz
                rows = plgpu.load(row_ref.at[pl.ds(i_start, block_dim)], mask=mask, other=0)
                cols = plgpu.load(col_ref.at[pl.ds(i_start, block_dim)], mask=mask, other=0)
                events = plgpu.load(
                    vector_ref.at[rows],
                    mask=mask,
                    other=False if vector_ref.dtype == jnp.bool_ else 0
                )

                if vector_ref.dtype == jnp.bool_:
                    event_mask = mask & events
                else:
                    event_mask = mask & (events > 0.)

                data = jnp.full((block_dim,), jnp.asarray(data_ref[0], dtype=posts_ref.dtype), dtype=posts_ref.dtype)
                atomic_add(posts_ref, cols, data, mask=event_mask)

        else:
            # coo.T @ v (heterogeneous weights)
            #
            # coo: [m, k]
            # v: [m]
            # result: [k]
            #
            def mv(
                data_ref,  # [nnz]
                row_ref,  # [nnz]
                col_ref,  # [nnz]
                vector_ref,  # [m]
                _,  # [k]
                posts_ref,  # [k]
            ):
                i = pl.program_id(0)
                i_start = i * block_dim
                mask = i_start + jnp.arange(block_dim) < nnz
                rows = plgpu.load(row_ref.at[pl.ds(i_start, block_dim)], mask=mask, other=0)
                cols = plgpu.load(col_ref.at[pl.ds(i_start, block_dim)], mask=mask, other=0)
                weights = plgpu.load(data_ref.at[pl.ds(i_start, block_dim)], mask=mask, other=0)
                events = plgpu.load(
                    vector_ref.at[rows],
                    mask=mask,
                    other=False if vector_ref.dtype == jnp.bool_ else 0
                )

                if vector_ref.dtype == jnp.bool_:
                    event_mask = mask & events
                else:
                    event_mask = mask & (events > 0.)

                data = jnp.asarray(weights, dtype=posts_ref.dtype)
                atomic_add(posts_ref, cols, data, mask=event_mask)

        def kernel(data, row, col, vector):
            fn = pl.pallas_call(
                mv,
                grid=(pl.cdiv(nnz, block_dim),),
                input_output_aliases={4: 0},
                out_shape=kwargs['outs'],
                backend='triton',
            )
            posts = jnp.zeros(kwargs['outs'][0].shape, dtype=kwargs['outs'][0].dtype)
            return fn(data, row, col, vector, posts)

        return kernel

    else:
        # coo @ v (non-transpose)
        # Also needs atomic operations due to COO race conditions
        if weight_info.size == 1:
            # coo @ v (homogeneous weights)
            #
            # coo: [m, k]
            # v: [k]
            # result: [m]
            #
            def mv(
                data_ref,  # [1]
                row_ref,  # [nnz]
                col_ref,  # [nnz]
                vector_ref,  # [k]
                _,  # [m]
                posts_ref,  # [m]
            ):
                i = pl.program_id(0)
                i_start = i * block_dim
                mask = i_start + jnp.arange(block_dim) < nnz
                rows = plgpu.load(row_ref.at[pl.ds(i_start, block_dim)], mask=mask, other=0)
                cols = plgpu.load(col_ref.at[pl.ds(i_start, block_dim)], mask=mask, other=0)
                events = plgpu.load(
                    vector_ref.at[cols],
                    mask=mask,
                    other=False if vector_ref.dtype == jnp.bool_ else 0
                )

                if vector_ref.dtype == jnp.bool_:
                    event_mask = mask & events
                else:
                    event_mask = mask & (events > 0.)

                data = jnp.full((block_dim,), jnp.asarray(data_ref[0], dtype=posts_ref.dtype), dtype=posts_ref.dtype)
                atomic_add(posts_ref, rows, data, mask=event_mask)

        else:
            # coo @ v (heterogeneous weights)
            #
            # coo: [m, k]
            # v: [k]
            # result: [m]
            #
            def mv(
                data_ref,  # [nnz]
                row_ref,  # [nnz]
                col_ref,  # [nnz]
                vector_ref,  # [k]
                _,  # [m]
                posts_ref,  # [m]
            ):
                i = pl.program_id(0)
                i_start = i * block_dim
                mask = i_start + jnp.arange(block_dim) < nnz
                rows = plgpu.load(row_ref.at[pl.ds(i_start, block_dim)], mask=mask, other=0)
                cols = plgpu.load(col_ref.at[pl.ds(i_start, block_dim)], mask=mask, other=0)
                weights = plgpu.load(data_ref.at[pl.ds(i_start, block_dim)], mask=mask, other=0)
                events = plgpu.load(
                    vector_ref.at[cols],
                    mask=mask,
                    other=False if vector_ref.dtype == jnp.bool_ else 0
                )

                if vector_ref.dtype == jnp.bool_:
                    event_mask = mask & events
                else:
                    event_mask = mask & (events > 0.)

                data = jnp.asarray(weights, dtype=posts_ref.dtype)
                atomic_add(posts_ref, rows, data, mask=event_mask)

        def kernel(data, row, col, vector):
            fn = pl.pallas_call(
                mv,
                grid=(pl.cdiv(nnz, block_dim),),
                input_output_aliases={4: 0},
                out_shape=kwargs['outs'],
                backend='triton',
            )
            posts = jnp.zeros(kwargs['outs'][0].shape, dtype=kwargs['outs'][0].dtype)
            return fn(data, row, col, vector, posts)

        return kernel


def _coomv_pallas_tpu_kernel(
    weight_info: jax.ShapeDtypeStruct,
    row_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    from jax.experimental import pallas as pl

    nnz = row_info.shape[0]
    block_dim = generate_block_dim(nnz)
    block_dim = 32 if block_dim < 32 else block_dim
    num_nnz_blocks = pl.cdiv(nnz, block_dim)
    out_len = kwargs['outs'][0].shape[0]

    if weight_info.size == 1:
        def mv(
            data_ref,  # [1]
            row_ref,  # [nnz]
            col_ref,  # [nnz]
            vector_ref,  # [m] or [k]
            _,  # [k] or [m]
            posts_ref,  # [k] or [m]
        ):
            i_out = pl.program_id(0)
            scalar_w = jnp.asarray(data_ref[0], dtype=posts_ref.dtype)

            def loop_fn(i_blk, acc):
                i_start = i_blk * block_dim
                elems = i_start + jnp.arange(block_dim)
                valid = elems < nnz
                safe_elems = jnp.where(valid, elems, 0)
                rows = row_ref[safe_elems]
                cols = col_ref[safe_elems]
                safe_rows = jnp.where(valid, rows, 0)
                safe_cols = jnp.where(valid, cols, 0)

                if transpose:
                    events = vector_ref[safe_rows]
                    out_idx = cols
                else:
                    events = vector_ref[safe_cols]
                    out_idx = rows

                if vector_ref.dtype == jnp.bool_:
                    active = events
                else:
                    active = events > 0.

                lane_mask = valid & (out_idx == i_out) & active
                return acc + jnp.sum(jnp.where(lane_mask, scalar_w, 0))

            acc0 = jnp.asarray(0, dtype=posts_ref.dtype)
            posts_ref[i_out] = jax.lax.fori_loop(0, num_nnz_blocks, loop_fn, acc0)

    else:
        def mv(
            data_ref,  # [nnz]
            row_ref,  # [nnz]
            col_ref,  # [nnz]
            vector_ref,  # [m] or [k]
            _,  # [k] or [m]
            posts_ref,  # [k] or [m]
        ):
            i_out = pl.program_id(0)

            def loop_fn(i_blk, acc):
                i_start = i_blk * block_dim
                elems = i_start + jnp.arange(block_dim)
                valid = elems < nnz
                safe_elems = jnp.where(valid, elems, 0)
                rows = row_ref[safe_elems]
                cols = col_ref[safe_elems]
                weights = jnp.asarray(data_ref[safe_elems], dtype=posts_ref.dtype)
                safe_rows = jnp.where(valid, rows, 0)
                safe_cols = jnp.where(valid, cols, 0)

                if transpose:
                    events = vector_ref[safe_rows]
                    out_idx = cols
                else:
                    events = vector_ref[safe_cols]
                    out_idx = rows

                if vector_ref.dtype == jnp.bool_:
                    active = events
                else:
                    active = events > 0.

                lane_mask = valid & (out_idx == i_out) & active
                return acc + jnp.sum(jnp.where(lane_mask, weights, 0))

            acc0 = jnp.asarray(0, dtype=posts_ref.dtype)
            posts_ref[i_out] = jax.lax.fori_loop(0, num_nnz_blocks, loop_fn, acc0)

    def kernel(data, row, col, vector):
        fn = pl.pallas_call(
            mv,
            grid=(out_len,),
            input_output_aliases={4: 0},
            out_shape=kwargs['outs'],
            backend='mosaic_tpu',
        )
        posts = jnp.zeros(kwargs['outs'][0].shape, dtype=kwargs['outs'][0].dtype)
        return fn(data, row, col, vector, posts)

    return kernel


def _coomv_jvp_vector(v_dot, data, row, col, v, *, shape, transpose, **kwargs):
    return [coomv(data, row, col, v_dot, shape=shape, transpose=transpose, backend=kwargs['backend'])]


def _coomv_jvp_weights(data_dot, data, row, col, v, *, shape, transpose, **kwargs):
    return binary_coomv_p_call(data_dot, row, col, v, shape=shape, transpose=transpose, backend=kwargs['backend'])


def _coomv_transpose_rule(ct, data, row, col, events, *, shape, transpose, **kwargs):
    assert not ad.is_undefined_primal(row)
    assert not ad.is_undefined_primal(col)
    ct = ct[0]

    if ad.is_undefined_primal(events):
        if type(ct) is ad.Zero:
            ct_events = ad.Zero(events)
        else:
            ct_events = coomv(
                data,
                row,
                col,
                ct,
                shape=shape,
                transpose=not transpose,
                backend=kwargs['backend']
            )
        return data, row, col, ct_events
    else:
        if type(ct) is ad.Zero:
            ct_values = ad.Zero(data)
        else:
            if data.aval.shape[0] == 1:  # scalar
                ct_values = binary_coomv_p_call(
                    jnp.ones(1, dtype=data.aval.dtype),
                    row,
                    col,
                    events,
                    shape=shape,
                    transpose=transpose,
                    backend=kwargs['backend']
                )[0]
                ct_values = jnp.inner(ct, ct_values).reshape(*data.aval.shape)
            else:
                ct_values = events[row] * ct[col] if transpose else events[col] * ct[row]
        return ct_values, row, col, events


def _coomv_batching(args, axes, **kwargs):
    if tuple(axes) == (None, None, None, 0):
        assert args[3].ndim == 2, 'Batching axis 0 requires 2D input.'
        r = binary_coomm_p_call(
            args[0],
            args[1],
            args[2],
            args[3].T,
            shape=kwargs['shape'],
            transpose=kwargs['transpose'],
            backend=kwargs['backend'],
        )
        return r, [1]

    elif tuple(axes) == (None, None, None, 1):
        assert args[3].ndim == 2, 'Batching axis 0 requires 2D input.'
        r = binary_coomm_p_call(
            args[0],
            args[1],
            args[2],
            args[3].T,
            shape=kwargs['shape'],
            transpose=kwargs['transpose'],
            backend=kwargs['backend'],
        )
        return r, [1]

    else:
        return general_batching_rule(binary_coomv_p_call, args, axes, **kwargs)


def _coomv_benchmark_data(*, platform):
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for transpose in (False, True):
        for homo in (True, False):
            for bool_event in (True, False):
                nnz = max(1, int(n_pre * n_post * prob))
                row = np.random.randint(0, n_pre, nnz, dtype=np.int32)
                col = np.random.randint(0, n_post, nnz, dtype=np.int32)
                weights = jnp.ones(1, dtype=dtype) if homo else jnp.ones(nnz, dtype=dtype)
                v_size = n_post if not transpose else n_pre
                if bool_event:
                    vector = jnp.asarray(np.random.rand(v_size) > 0.5, dtype=jnp.bool_)
                else:
                    vector = jnp.asarray(np.random.rand(v_size), dtype=dtype)
                name = f"{'T' if transpose else 'NT'},{'homo' if homo else 'hetero'},{'bool' if bool_event else 'float'}"
                configs.append(
                    BenchmarkConfig(
                        name,
                        (weights, jnp.asarray(row), jnp.asarray(col), vector),
                        {'shape': (n_pre, n_post), 'transpose': transpose}
                    )
                )
    return configs


def binary_coomv_p_call(
    weights,
    row,
    col,
    v,
    *,
    shape: Sequence[int],
    transpose: bool,
    backend: Optional[str] = None,
):
    """
    Low-level primitive call for event-driven COO sparse matrix-vector multiplication.

    Validates inputs, constructs output metadata, and dispatches to the
    registered ``XLACustomKernel`` (``binary_coomv_p``) which selects a
    backend-specific kernel (Numba, Warp, Pallas GPU, or Pallas TPU).

    Unlike :func:`binary_coomv`, this function does **not** handle physical
    units and returns a raw list of JAX arrays.

    Parameters
    ----------
    weights : jax.Array
        Non-zero values of the sparse matrix.  Must be a floating-point
        scalar (shape ``(1,)``) for homogeneous weights or a 1-D array
        of length ``nnz`` for heterogeneous weights.
    row : jax.Array
        1-D integer array of row indices with length ``nnz``.
    col : jax.Array
        1-D integer array of column indices with length ``nnz``.
    v : jax.Array
        Dense event vector.  Entries are treated as active when ``True``
        (boolean) or ``> 0`` (numeric).  Shape must be ``(shape[0],)``
        when ``transpose=True`` or ``(shape[1],)`` when
        ``transpose=False``.
    shape : Sequence[int]
        The ``(m, k)`` shape of the logical sparse matrix.
    transpose : bool
        If ``True``, compute ``A^T @ v`` instead of ``A @ v``.
    backend : str or None, optional
        Compute backend override (``'numba'``, ``'warp'``, ``'pallas'``).
        When ``None`` the backend is selected automatically.

    Returns
    -------
    list of jax.Array
        A single-element list containing the result vector.  Shape is
        ``(shape[0],)`` when ``transpose=False`` or ``(shape[1],)`` when
        ``transpose=True``, with dtype matching ``weights``.

    Raises
    ------
    ValueError
        If ``row`` or ``col`` are not 1-D arrays, have different lengths,
        ``v`` is not 1-D, ``weights`` is not scalar or 1-D, weights
        length is neither 1 nor ``nnz``, or ``v`` length is incompatible
        with ``shape`` and ``transpose``.
    AssertionError
        If ``weights`` dtype is not a floating-point type.

    See Also
    --------
    binary_coomv : High-level wrapper with physical unit support.

    Notes
    -----
    When ``nnz == 0`` the function short-circuits and returns a zero vector
    without dispatching to any kernel.

    The function registers JVP rules, transpose rules, and batching rules
    on the underlying ``binary_coomv_p`` primitive so that it integrates
    with JAX's autodiff and ``vmap`` transformations.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._coo.binary import binary_coomv_p_call
        >>> weights = jnp.array([1.0, 2.0, 3.0])
        >>> row = jnp.array([0, 1, 2])
        >>> col = jnp.array([2, 0, 1])
        >>> v = jnp.array([True, False, True])
        >>> result = binary_coomv_p_call(
        ...     weights, row, col, v, shape=(3, 3), transpose=False
        ... )
        >>> result[0]  # the output vector
    """
    row = jnp.asarray(row)
    col = jnp.asarray(col)
    v = jnp.asarray(v)
    if row.ndim != 1 or col.ndim != 1:
        raise ValueError(f'`row` and `col` must be 1D arrays, got row.ndim={row.ndim}, col.ndim={col.ndim}.')
    if row.shape[0] != col.shape[0]:
        raise ValueError(f'`row` and `col` must have the same length, got {row.shape[0]} and {col.shape[0]}.')
    if v.ndim != 1:
        raise ValueError(f'`v` must be a 1D array, got ndim={v.ndim}.')

    if jnp.ndim(weights) == 0:
        weights = jnp.asarray([weights])
    else:
        weights = jnp.asarray(weights)
    if weights.ndim != 1:
        raise ValueError(f'`weights` must be a scalar or 1D array, got ndim={weights.ndim}.')
    assert jnp.issubdtype(weights.dtype, jnp.floating), 'Weights must be a floating-point type.'

    nnz = row.shape[0]
    if weights.shape[0] not in (1, nnz):
        raise ValueError(f'`weights` length must be 1 or nnz={nnz}, got {weights.shape[0]}.')

    expected_v = shape[0] if transpose else shape[1]
    if v.shape[0] != expected_v:
        raise ValueError(f'`v` has incompatible length {v.shape[0]} for shape={tuple(shape)}, transpose={transpose}.')

    out_len = shape[1] if transpose else shape[0]
    if nnz == 0:
        return [jnp.zeros((out_len,), dtype=weights.dtype)]

    out_info = jax.ShapeDtypeStruct([out_len], weights.dtype)

    # Call the custom kernel with the provided arguments and output information
    return binary_coomv_p(
        weights,
        row,
        col,
        v,
        outs=[out_info],
        shape=shape,
        transpose=transpose,
        backend=backend,
        # Provide shape and dtype information for row indices
        row_info=jax.ShapeDtypeStruct(row.shape, row.dtype),
        # Provide shape and dtype information for column indices
        col_info=jax.ShapeDtypeStruct(col.shape, col.dtype),
        # Provide shape and dtype information for non-zero values
        weight_info=jax.ShapeDtypeStruct(weights.shape, weights.dtype),
        # Provide shape and dtype information for the dense vector
        vector_info=jax.ShapeDtypeStruct(v.shape, v.dtype),
    )


binary_coomv_p = XLACustomKernel(
    'binary_coomv',
    doc="""
Low-level XLA custom-kernel primitive for ``binary_coomv``.

This ``XLACustomKernel`` instance dispatches the binary (event-driven) COO
sparse matrix-vector multiplication operation to registered backends
(``numba``, ``warp``, ``pallas``), using runtime shape/dtype metadata
provided by the high-level wrapper.

The operation computes ``result[i] = sum_j A[i, j] * (v[j] > 0)`` when
``transpose=False`` or ``result[j] = sum_i A[i, j] * (v[i] > 0)`` when
``transpose=True``, where only active (nonzero) events in ``v`` contribute
to the output. This makes the operation efficient for spike-based neural
network simulations.

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``binary_coomv_p.available_backends(platform)``,
and the default backend can be configured with ``binary_coomv_p.set_default(platform, backend)``.

See Also
--------
binary_coomv : High-level user-facing function wrapper.
"""
)
binary_coomv_p.def_numba_kernel(_coomv_numba_kernel)
binary_coomv_p.def_warp_kernel(_coomv_warp_kernel)
binary_coomv_p.def_pallas_kernel('gpu', _coomv_pallas_gpu_kernel)
binary_coomv_p.def_pallas_kernel('tpu', _coomv_pallas_tpu_kernel)
binary_coomv_p.def_jvp_rule2(_coomv_jvp_weights, None, None, _coomv_jvp_vector)
binary_coomv_p.def_transpose_rule(_coomv_transpose_rule)
binary_coomv_p.def_batching_rule(_coomv_batching)
binary_coomv_p.def_call(binary_coomv_p_call)
binary_coomv_p.def_tags('coo', 'binary')
binary_coomv_p.def_benchmark_data(_coomv_benchmark_data)


# =============================================================================
# COO Matrix-Matrix Multiplication (coomm)
# =============================================================================


def _coomm_numba_kernel(
    weight_info: jax.ShapeDtypeStruct,
    matrix_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    import numba

    if transpose:
        if weight_info.size == 1:
            # transpose=True, homogeneous
            if matrix_info.dtype == jnp.bool_:
                # bool
                @numba.njit(fastmath=True)
                def mm(weights, row, col, B, posts):
                    posts[:] = 0.
                    w = weights[0]
                    for i in range(row.shape[0]):
                        for j in range(B.shape[1]):
                            if B[row[i], j]:
                                posts[col[i], j] += w
            else:
                @numba.njit(fastmath=True)
                def mm(weights, row, col, B, posts):
                    posts[:] = 0.
                    w = weights[0]
                    for i in range(row.shape[0]):
                        for j in range(B.shape[1]):
                            if B[row[i], j] > 0.:
                                posts[col[i], j] += w
        else:
            # transpose=True, heterogeneous
            if matrix_info.dtype == jnp.bool_:
                # bool
                @numba.njit(fastmath=True)
                def mm(weights, row, col, B, posts):
                    posts[:] = 0.
                    for i in range(row.shape[0]):
                        for j in range(B.shape[1]):
                            if B[row[i], j]:
                                posts[col[i], j] += weights[i]
            else:
                @numba.njit(fastmath=True)
                def mm(weights, row, col, B, posts):
                    posts[:] = 0.
                    for i in range(row.shape[0]):
                        for j in range(B.shape[1]):
                            if B[row[i], j] > 0.:
                                posts[col[i], j] += weights[i]
    else:
        if weight_info.size == 1:
            # transpose=False, homogeneous
            if matrix_info.dtype == jnp.bool_:
                # bool
                @numba.njit(fastmath=True)
                def mm(weights, row, col, B, posts):
                    posts[:] = 0.
                    w = weights[0]
                    for i in range(row.shape[0]):
                        for j in range(B.shape[1]):
                            if B[col[i], j]:
                                posts[row[i], j] += w
            else:
                @numba.njit(fastmath=True)
                def mm(weights, row, col, B, posts):
                    posts[:] = 0.
                    w = weights[0]
                    for i in range(row.shape[0]):
                        for j in range(B.shape[1]):
                            if B[col[i], j] > 0.:
                                posts[row[i], j] += w
        else:
            # transpose=False, heterogeneous
            if matrix_info.dtype == jnp.bool_:
                # bool
                @numba.njit(fastmath=True)
                def mm(weights, row, col, B, posts):
                    posts[:] = 0.
                    for i in range(row.shape[0]):
                        for j in range(B.shape[1]):
                            if B[col[i], j]:
                                posts[row[i], j] += weights[i]
            else:
                @numba.njit(fastmath=True)
                def mm(weights, row, col, B, posts):
                    posts[:] = 0.
                    for i in range(row.shape[0]):
                        for j in range(B.shape[1]):
                            if B[col[i], j] > 0.:
                                posts[row[i], j] += weights[i]

    def kernel(weights, row, col, B):
        return numba_kernel(mm, outs=kwargs['outs'])(weights, row, col, B)

    return kernel


def _coomm_warp_kernel(
    weight_info: jax.ShapeDtypeStruct,
    matrix_info: jax.ShapeDtypeStruct,
    row_info: jax.ShapeDtypeStruct,
    col_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    import warp
    from warp.jax_experimental import jax_kernel

    weight_warp_info = jaxinfo_to_warpinfo(weight_info)
    row_warp_info = jaxinfo_to_warpinfo(row_info)
    col_warp_info = jaxinfo_to_warpinfo(col_info)
    spike_warp_info = jaxinfo_to_warpinfo(matrix_info)
    out_warp_info = jaxinfo_to_warpinfo(kwargs['outs'][0])

    if transpose:
        if weight_info.size == 1:
            # transpose=True, homogeneous
            if matrix_info.dtype == jnp.bool_:
                # bool
                @warp.kernel
                def mm(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    B: spike_warp_info,
                    posts: out_warp_info
                ):
                    i, j = warp.tid()
                    w = weights[0]
                    if B[row[i], j]:
                        warp.atomic_add(posts, col[i], j, w)
            else:
                @warp.kernel
                def mm(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    B: spike_warp_info,
                    posts: out_warp_info
                ):
                    i, j = warp.tid()
                    w = weights[0]
                    if B[row[i], j] > 0.:
                        warp.atomic_add(posts, col[i], j, w)
        else:
            # transpose=True, heterogeneous
            if matrix_info.dtype == jnp.bool_:
                # bool
                @warp.kernel
                def mm(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    B: spike_warp_info,
                    posts: out_warp_info
                ):
                    i, j = warp.tid()
                    if B[row[i], j]:
                        warp.atomic_add(posts, col[i], j, weights[i])
            else:
                @warp.kernel
                def mm(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    B: spike_warp_info,
                    posts: out_warp_info
                ):
                    i, j = warp.tid()
                    if B[row[i], j] > 0.:
                        warp.atomic_add(posts, col[i], j, weights[i])
    else:
        if weight_info.size == 1:
            # transpose=False, homogeneous
            if matrix_info.dtype == jnp.bool_:
                # bool
                @warp.kernel
                def mm(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    B: spike_warp_info,
                    posts: out_warp_info
                ):
                    i, j = warp.tid()
                    w = weights[0]
                    if B[col[i], j]:
                        warp.atomic_add(posts, row[i], j, w)
            else:
                @warp.kernel
                def mm(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    B: spike_warp_info,
                    posts: out_warp_info
                ):
                    i, j = warp.tid()
                    w = weights[0]
                    if B[col[i], j] > 0.:
                        warp.atomic_add(posts, row[i], j, w)
        else:
            # transpose=False, heterogeneous
            if matrix_info.dtype == jnp.bool_:
                # bool
                @warp.kernel
                def mm(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    B: spike_warp_info,
                    posts: out_warp_info
                ):
                    i, j = warp.tid()
                    if B[col[i], j]:
                        warp.atomic_add(posts, row[i], j, weights[i])
            else:
                @warp.kernel
                def mm(
                    weights: weight_warp_info,
                    row: row_warp_info,
                    col: col_warp_info,
                    B: spike_warp_info,
                    posts: out_warp_info
                ):
                    i, j = warp.tid()
                    if B[col[i], j] > 0.:
                        warp.atomic_add(posts, row[i], j, weights[i])

    def kernel(weights, row, col, B):
        dim = (row_info.shape[0], matrix_info.shape[1])
        out_info = kwargs['outs'][0]
        fn = jax_kernel(mm, launch_dims=dim, num_outputs=1, in_out_argnames=['posts'])
        return fn(weights, row, col, B, jnp.zeros(out_info.shape, out_info.dtype))

    return kernel


def _coomm_pallas_gpu_kernel(
    weight_info: jax.ShapeDtypeStruct,
    matrix_info: jax.ShapeDtypeStruct,
    row_info: jax.ShapeDtypeStruct,
    transpose: bool,
    **kwargs
):
    from jax.experimental import pallas as pl
    from jax.experimental.pallas import triton as plgpu

    nnz = row_info.shape[0]
    n = matrix_info.shape[1]
    block_dim = generate_block_dim(nnz)
    block_dim = 32 if block_dim < 32 else block_dim
    block_dim_n = generate_block_dim(n, 512)

    if transpose:
        if weight_info.size == 1:
            # coo.T @ B (homogeneous weights)
            #
            # coo: [m, k]
            # B: [m, n]
            # result: [k, n]
            #
            def mm(
                data_ref,  # [1]
                row_ref,  # [nnz]
                col_ref,  # [nnz]
                B_ref,  # [m, n]
                _,  # [k, n]
                posts_ref,  # [k, n]
            ):
                i = pl.program_id(0)
                i_n = pl.program_id(1)
                i_col_start = i_n * block_dim_n
                col_mask = (i_col_start + jnp.arange(block_dim_n)) < B_ref.shape[1]

                i_start = i * block_dim
                scalar_w = jnp.asarray(data_ref[0], dtype=posts_ref.dtype)

                def loop_fn(idx, _):
                    elem = i_start + idx
                    valid = elem < nnz
                    row_idx = plgpu.load(row_ref.at[elem], mask=valid, other=0)
                    col_idx = plgpu.load(col_ref.at[elem], mask=valid, other=0)
                    events = plgpu.load(
                        B_ref.at[row_idx, pl.ds(i_col_start, block_dim_n)],
                        mask=col_mask,
                        other=False if B_ref.dtype == jnp.bool_ else 0
                    )
                    if B_ref.dtype == jnp.bool_:
                        event_mask = col_mask & events & valid
                    else:
                        event_mask = col_mask & (events > 0.) & valid
                    data = jnp.full((block_dim_n,), scalar_w, dtype=posts_ref.dtype)
                    plgpu.atomic_add(posts_ref, (col_idx, pl.ds(i_col_start, block_dim_n)), data, mask=event_mask)

                jax.lax.fori_loop(0, block_dim, loop_fn, None)

        else:
            # coo.T @ B (heterogeneous weights)
            #
            # coo: [m, k]
            # B: [m, n]
            # result: [k, n]
            #
            def mm(
                data_ref,  # [nnz]
                row_ref,  # [nnz]
                col_ref,  # [nnz]
                B_ref,  # [m, n]
                _,  # [k, n]
                posts_ref,  # [k, n]
            ):
                i = pl.program_id(0)
                i_n = pl.program_id(1)
                i_col_start = i_n * block_dim_n
                col_mask = (i_col_start + jnp.arange(block_dim_n)) < B_ref.shape[1]

                i_start = i * block_dim

                def loop_fn(idx, _):
                    elem = i_start + idx
                    valid = elem < nnz
                    row_idx = plgpu.load(row_ref.at[elem], mask=valid, other=0)
                    col_idx = plgpu.load(col_ref.at[elem], mask=valid, other=0)
                    w = jnp.asarray(plgpu.load(data_ref.at[elem], mask=valid, other=0), dtype=posts_ref.dtype)
                    events = plgpu.load(
                        B_ref.at[row_idx, pl.ds(i_col_start, block_dim_n)],
                        mask=col_mask,
                        other=False if B_ref.dtype == jnp.bool_ else 0
                    )
                    if B_ref.dtype == jnp.bool_:
                        event_mask = col_mask & events & valid
                    else:
                        event_mask = col_mask & (events > 0.) & valid
                    data = jnp.full((block_dim_n,), w, dtype=posts_ref.dtype)
                    plgpu.atomic_add(posts_ref, (col_idx, pl.ds(i_col_start, block_dim_n)), data, mask=event_mask)

                jax.lax.fori_loop(0, block_dim, loop_fn, None)

        def kernel(data, row, col, B):
            fn = pl.pallas_call(
                mm,
                grid=(pl.cdiv(nnz, block_dim), pl.cdiv(n, block_dim_n)),
                input_output_aliases={4: 0},
                out_shape=kwargs['outs'],
                backend='triton',
            )
            posts = jnp.zeros(kwargs['outs'][0].shape, dtype=kwargs['outs'][0].dtype)
            return fn(data, row, col, B, posts)

        return kernel

    else:
        # coo @ B (non-transpose)
        if weight_info.size == 1:
            # coo @ B (homogeneous weights)
            #
            # coo: [m, k]
            # B: [k, n]
            # result: [m, n]
            #
            def mm(
                data_ref,  # [1]
                row_ref,  # [nnz]
                col_ref,  # [nnz]
                B_ref,  # [k, n]
                _,  # [m, n]
                posts_ref,  # [m, n]
            ):
                i = pl.program_id(0)
                i_n = pl.program_id(1)
                i_col_start = i_n * block_dim_n
                col_mask = (i_col_start + jnp.arange(block_dim_n)) < B_ref.shape[1]

                i_start = i * block_dim
                scalar_w = jnp.asarray(data_ref[0], dtype=posts_ref.dtype)

                def loop_fn(idx, _):
                    elem = i_start + idx
                    valid = elem < nnz
                    row_idx = plgpu.load(row_ref.at[elem], mask=valid, other=0)
                    col_idx = plgpu.load(col_ref.at[elem], mask=valid, other=0)
                    events = plgpu.load(
                        B_ref.at[col_idx, pl.ds(i_col_start, block_dim_n)],
                        mask=col_mask,
                        other=False if B_ref.dtype == jnp.bool_ else 0
                    )
                    if B_ref.dtype == jnp.bool_:
                        event_mask = col_mask & events & valid
                    else:
                        event_mask = col_mask & (events > 0.) & valid
                    data = jnp.full((block_dim_n,), scalar_w, dtype=posts_ref.dtype)
                    plgpu.atomic_add(posts_ref, (row_idx, pl.ds(i_col_start, block_dim_n)), data, mask=event_mask)

                jax.lax.fori_loop(0, block_dim, loop_fn, None)

        else:
            # coo @ B (heterogeneous weights)
            #
            # coo: [m, k]
            # B: [k, n]
            # result: [m, n]
            #
            def mm(
                data_ref,  # [nnz]
                row_ref,  # [nnz]
                col_ref,  # [nnz]
                B_ref,  # [k, n]
                _,  # [m, n]
                posts_ref,  # [m, n]
            ):
                i = pl.program_id(0)
                i_n = pl.program_id(1)
                i_col_start = i_n * block_dim_n
                col_mask = (i_col_start + jnp.arange(block_dim_n)) < B_ref.shape[1]

                i_start = i * block_dim

                def loop_fn(idx, _):
                    elem = i_start + idx
                    valid = elem < nnz
                    row_idx = plgpu.load(row_ref.at[elem], mask=valid, other=0)
                    col_idx = plgpu.load(col_ref.at[elem], mask=valid, other=0)
                    w = jnp.asarray(plgpu.load(data_ref.at[elem], mask=valid, other=0), dtype=posts_ref.dtype)
                    events = plgpu.load(
                        B_ref.at[col_idx, pl.ds(i_col_start, block_dim_n)],
                        mask=col_mask,
                        other=False if B_ref.dtype == jnp.bool_ else 0
                    )
                    if B_ref.dtype == jnp.bool_:
                        event_mask = col_mask & events & valid
                    else:
                        event_mask = col_mask & (events > 0.) & valid
                    data = jnp.full((block_dim_n,), w, dtype=posts_ref.dtype)
                    plgpu.atomic_add(posts_ref, (row_idx, pl.ds(i_col_start, block_dim_n)), data, mask=event_mask)

                jax.lax.fori_loop(0, block_dim, loop_fn, None)

        def kernel(data, row, col, B):
            fn = pl.pallas_call(
                mm,
                grid=(pl.cdiv(nnz, block_dim), pl.cdiv(n, block_dim_n)),
                input_output_aliases={4: 0},
                out_shape=kwargs['outs'],
                backend='triton',
            )
            posts = jnp.zeros(kwargs['outs'][0].shape, dtype=kwargs['outs'][0].dtype)
            return fn(data, row, col, B, posts)

        return kernel


def _coomm_pallas_tpu_kernel(
    weight_info: jax.ShapeDtypeStruct,
    matrix_info: jax.ShapeDtypeStruct,
    row_info: jax.ShapeDtypeStruct,
    shape: MatrixShape,
    transpose: bool,
    **kwargs
):
    from jax.experimental import pallas as pl
    from jax.experimental.pallas import tpu as pltpu

    nnz = row_info.shape[0]
    n = matrix_info.shape[1]
    block_dim = generate_block_dim(nnz)
    block_dim = 32 if block_dim < 32 else block_dim
    block_dim_n = generate_block_dim(n, 512)
    num_nnz_blocks = pl.cdiv(nnz, block_dim)
    out_rows = kwargs['outs'][0].shape[0]

    if weight_info.size == 1:
        def mm(
            data_ref,  # [1]
            row_ref,  # [nnz]
            col_ref,  # [nnz]
            B_ref,  # [m, n] or [k, n]
            _,  # [k, n] or [m, n]
            posts_ref,  # [k, n] or [m, n]
        ):
            i_out = pl.program_id(0)
            i_n = pl.program_id(1)
            i_col_start = i_n * block_dim_n
            col_mask = (i_col_start + jnp.arange(block_dim_n)) < B_ref.shape[1]
            scalar_w = jnp.asarray(data_ref[0], dtype=posts_ref.dtype)

            def loop_fn(i_blk, acc):
                i_start = i_blk * block_dim
                elems = i_start + jnp.arange(block_dim)
                valid = elems < nnz
                safe_elems = jnp.where(valid, elems, 0)
                rows = row_ref[safe_elems]
                cols = col_ref[safe_elems]
                safe_rows = jnp.where(valid, rows, 0)
                safe_cols = jnp.where(valid, cols, 0)

                if transpose:
                    events = B_ref[safe_rows, pl.ds(i_col_start, block_dim_n)]
                    out_idx = cols
                else:
                    events = B_ref[safe_cols, pl.ds(i_col_start, block_dim_n)]
                    out_idx = rows

                if B_ref.dtype == jnp.bool_:
                    active = events
                else:
                    active = events > 0.

                lane_mask = valid & (out_idx == i_out)
                active_mask = lane_mask[:, None] & active & col_mask[None, :]
                return acc + jnp.sum(jnp.where(active_mask, scalar_w, 0), axis=0)

            acc0 = jnp.zeros((block_dim_n,), dtype=posts_ref.dtype)
            acc = jax.lax.fori_loop(0, num_nnz_blocks, loop_fn, acc0)
            pltpu.store(posts_ref.at[i_out, pl.ds(i_col_start, block_dim_n)], acc, mask=col_mask)

    else:
        def mm(
            data_ref,  # [nnz]
            row_ref,  # [nnz]
            col_ref,  # [nnz]
            B_ref,  # [m, n] or [k, n]
            _,  # [k, n] or [m, n]
            posts_ref,  # [k, n] or [m, n]
        ):
            i_out = pl.program_id(0)
            i_n = pl.program_id(1)
            i_col_start = i_n * block_dim_n
            col_mask = (i_col_start + jnp.arange(block_dim_n)) < B_ref.shape[1]

            def loop_fn(i_blk, acc):
                i_start = i_blk * block_dim
                elems = i_start + jnp.arange(block_dim)
                valid = elems < nnz
                safe_elems = jnp.where(valid, elems, 0)
                rows = row_ref[safe_elems]
                cols = col_ref[safe_elems]
                weights = jnp.asarray(data_ref[safe_elems], dtype=posts_ref.dtype)
                safe_rows = jnp.where(valid, rows, 0)
                safe_cols = jnp.where(valid, cols, 0)

                if transpose:
                    events = B_ref[safe_rows, pl.ds(i_col_start, block_dim_n)]
                    out_idx = cols
                else:
                    events = B_ref[safe_cols, pl.ds(i_col_start, block_dim_n)]
                    out_idx = rows

                if B_ref.dtype == jnp.bool_:
                    active = events
                else:
                    active = events > 0.

                lane_mask = valid & (out_idx == i_out)
                active_mask = lane_mask[:, None] & active & col_mask[None, :]
                weighted = jnp.where(active_mask, weights[:, None], 0)
                return acc + jnp.sum(weighted, axis=0)

            acc0 = jnp.zeros((block_dim_n,), dtype=posts_ref.dtype)
            acc = jax.lax.fori_loop(0, num_nnz_blocks, loop_fn, acc0)
            pltpu.store(posts_ref.at[i_out, pl.ds(i_col_start, block_dim_n)], acc, mask=col_mask)

    def kernel(data, row, col, B):
        fn = pl.pallas_call(
            mm,
            grid=(out_rows, pl.cdiv(n, block_dim_n)),
            input_output_aliases={4: 0},
            out_shape=kwargs['outs'],
            backend='mosaic_tpu',
        )
        posts = jnp.zeros(kwargs['outs'][0].shape, dtype=kwargs['outs'][0].dtype)
        return fn(data, row, col, B, posts)

    return kernel


def _coomm_jvp_left(data_dot, data, row, col, B, *, shape, transpose, **kwargs):
    return [coomm(data_dot, row, col, B, shape=shape, transpose=transpose, backend=kwargs['backend'])]


def _coomm_jvp_right(B_dot, data, row, col, B, *, shape, transpose, **kwargs):
    return [coomm(data, row, col, B_dot, shape=shape, transpose=transpose, backend=kwargs['backend'])]


def _coomm_transpose_rule(ct, data, row, col, B, *, shape, transpose, **kwargs):
    assert not ad.is_undefined_primal(row)
    assert not ad.is_undefined_primal(col)
    ct = ct[0]

    if ad.is_undefined_primal(B):
        if type(ct) is ad.Zero:
            dB = ad.Zero(B)
        else:
            dB = coomm(data, row, col, ct, shape=shape, transpose=not transpose, backend=kwargs['backend'])
        return data, row, col, dB
    else:
        if type(ct) is ad.Zero:
            d_data = ad.Zero(data)
        else:
            d_data = sddmm_coo_indices(ct, B, row, col).data
        return d_data, row, col, B


def _coomm_batching(args, axes, **kwargs):
    if tuple(axes) == (None, None, None, 0):
        assert args[3].ndim == 3, 'Batching axis 0 requires 3D input.'
        batch_size, m, n = args[3].shape
        B = jnp.transpose(args[3], (1, 0, 2)).reshape(m, batch_size * n)
        r = binary_coomm_p_call(
            args[0],
            args[1],
            args[2],
            B,
            shape=kwargs['shape'],
            transpose=kwargs['transpose'],
            backend=kwargs['backend'],
        )
        r = jnp.reshape(r[0], [r[0].shape[0], batch_size, n])
        return [r], [1]

    elif tuple(axes) == (None, None, None, 1):
        assert args[3].ndim == 3, 'Batching axis 0 requires 3D input.'
        m, batch_size, n = args[3].shape
        B = args[3].reshape(m, batch_size * n)
        r = binary_coomm_p_call(
            args[0],
            args[1],
            args[2],
            B,
            shape=kwargs['shape'],
            transpose=kwargs['transpose'],
            backend=kwargs['backend'],
        )
        r = jnp.reshape(r[0], [r[0].shape[0], batch_size, n])
        return [r], [1]

    elif tuple(axes) == (None, None, None, 2):
        assert args[3].ndim == 3, 'Batching axis 0 requires 3D input.'
        m, n, batch_size = args[3].shape
        B = args[3].reshape(m, batch_size * n)
        r = binary_coomm_p_call(
            args[0],
            args[1],
            args[2],
            B,
            shape=kwargs['shape'],
            transpose=kwargs['transpose'],
            backend=kwargs['backend'],
        )
        r = jnp.reshape(r[0], [r[0].shape[0], n, batch_size])
        return [r], [2]

    else:
        return general_batching_rule(binary_coomm_p_call, args, axes, **kwargs)


def _coomm_benchmark_data(*, platform):
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for transpose in (False, True):
        for homo in (True, False):
            for bool_event in (True, False):
                nnz = max(1, int(n_pre * n_post * prob))
                row = np.random.randint(0, n_pre, nnz, dtype=np.int32)
                col = np.random.randint(0, n_post, nnz, dtype=np.int32)
                weights = jnp.ones(1, dtype=dtype) if homo else jnp.ones(nnz, dtype=dtype)
                b_rows = n_post if not transpose else n_pre
                if bool_event:
                    B = jnp.asarray(np.random.rand(b_rows, 10) > 0.5, dtype=jnp.bool_)
                else:
                    B = jnp.asarray(np.random.rand(b_rows, 10), dtype=dtype)
                name = f"{'T' if transpose else 'NT'},{'homo' if homo else 'hetero'},{'bool' if bool_event else 'float'}"
                configs.append(
                    BenchmarkConfig(
                        name,
                        (weights, jnp.asarray(row), jnp.asarray(col), B),
                        {'shape': (n_pre, n_post), 'transpose': transpose}
                    )
                )
    return configs


def binary_coomm_p_call(
    weights,
    row,
    col,
    B,
    *,
    shape: Sequence[int],
    transpose: bool,
    backend: Optional[str] = None,
):
    """
    Low-level primitive call for event-driven COO sparse matrix-matrix multiplication.

    Validates inputs, constructs output metadata, and dispatches to the
    registered ``XLACustomKernel`` (``binary_coomm_p``) which selects a
    backend-specific kernel (Numba, Warp, Pallas GPU, or Pallas TPU).

    Unlike :func:`binary_coomm`, this function does **not** handle physical
    units and returns a raw list of JAX arrays.

    Parameters
    ----------
    weights : jax.Array
        Non-zero values of the sparse matrix.  Must be a floating-point
        scalar (shape ``(1,)``) for homogeneous weights or a 1-D array
        of length ``nnz`` for heterogeneous weights.
    row : jax.Array
        1-D integer array of row indices with length ``nnz``.
    col : jax.Array
        1-D integer array of column indices with length ``nnz``.
    B : jax.Array
        Dense event matrix of shape ``(shape[0], n)`` when
        ``transpose=True`` or ``(shape[1], n)`` when ``transpose=False``.
        Entries are treated as active when ``True`` (boolean) or ``> 0``
        (numeric).
    shape : Sequence[int]
        The ``(m, k)`` shape of the logical sparse matrix.
    transpose : bool
        If ``True``, compute ``A^T @ B`` instead of ``A @ B``.
    backend : str or None, optional
        Compute backend override (``'numba'``, ``'warp'``, ``'pallas'``).
        When ``None`` the backend is selected automatically.

    Returns
    -------
    list of jax.Array
        A single-element list containing the result matrix.  Shape is
        ``(shape[0], n)`` when ``transpose=False`` or ``(shape[1], n)``
        when ``transpose=True``, with dtype matching ``weights``.

    Raises
    ------
    ValueError
        If ``row`` or ``col`` are not 1-D arrays, have different lengths,
        ``B`` is not 2-D, ``weights`` is not scalar or 1-D, weights
        length is neither 1 nor ``nnz``, or ``B`` row count is
        incompatible with ``shape`` and ``transpose``.
    AssertionError
        If ``weights`` dtype is not a floating-point type.

    See Also
    --------
    binary_coomm : High-level wrapper with physical unit support.

    Notes
    -----
    When ``nnz == 0`` the function short-circuits and returns a zero
    matrix without dispatching to any kernel.

    The function registers JVP rules, transpose rules, and batching rules
    on the underlying ``binary_coomm_p`` primitive so that it integrates
    with JAX's autodiff and ``vmap`` transformations.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._coo.binary import binary_coomm_p_call
        >>> weights = jnp.array([1.0, 2.0, 3.0])
        >>> row = jnp.array([0, 1, 2])
        >>> col = jnp.array([2, 0, 1])
        >>> B = jnp.array([[True, False],
        ...                [False, True],
        ...                [True, True]])
        >>> result = binary_coomm_p_call(
        ...     weights, row, col, B, shape=(3, 3), transpose=False
        ... )
        >>> result[0]  # the output matrix
    """
    row = jnp.asarray(row)
    col = jnp.asarray(col)
    B = jnp.asarray(B)
    if row.ndim != 1 or col.ndim != 1:
        raise ValueError(f'`row` and `col` must be 1D arrays, got row.ndim={row.ndim}, col.ndim={col.ndim}.')
    if row.shape[0] != col.shape[0]:
        raise ValueError(f'`row` and `col` must have the same length, got {row.shape[0]} and {col.shape[0]}.')
    if B.ndim != 2:
        raise ValueError(f'`B` must be a 2D array, got ndim={B.ndim}.')

    if jnp.ndim(weights) == 0:
        weights = jnp.asarray([weights])
    else:
        weights = jnp.asarray(weights)
    if weights.ndim != 1:
        raise ValueError(f'`weights` must be a scalar or 1D array, got ndim={weights.ndim}.')
    assert jnp.issubdtype(weights.dtype, jnp.floating), 'Weights must be a floating-point type.'

    nnz = row.shape[0]
    if weights.shape[0] not in (1, nnz):
        raise ValueError(f'`weights` length must be 1 or nnz={nnz}, got {weights.shape[0]}.')

    expected_b_rows = shape[0] if transpose else shape[1]
    if B.shape[0] != expected_b_rows:
        raise ValueError(f'`B` has incompatible shape {B.shape} for shape={tuple(shape)}, transpose={transpose}.')

    out_rows = shape[1] if transpose else shape[0]
    if nnz == 0:
        return [jnp.zeros((out_rows, B.shape[1]), dtype=weights.dtype)]

    out_info = jax.ShapeDtypeStruct([out_rows, B.shape[1]], weights.dtype)
    # Call the custom kernel with the provided arguments and output information
    return binary_coomm_p(
        weights,
        row,
        col,
        B,
        outs=[out_info],
        shape=shape,
        transpose=transpose,
        backend=backend,
        row_info=jax.ShapeDtypeStruct(row.shape, row.dtype),
        col_info=jax.ShapeDtypeStruct(col.shape, col.dtype),
        weight_info=jax.ShapeDtypeStruct(weights.shape, weights.dtype),
        matrix_info=jax.ShapeDtypeStruct(B.shape, B.dtype),
    )


binary_coomm_p = XLACustomKernel(
    'binary_coomm',
    doc="""
Low-level XLA custom-kernel primitive for ``binary_coomm``.

This ``XLACustomKernel`` instance dispatches the binary (event-driven) COO
sparse matrix-matrix multiplication operation to registered backends
(``numba``, ``warp``, ``pallas``), using runtime shape/dtype metadata
provided by the high-level wrapper.

The operation computes ``result[i, n] = sum_j A[i, j] * (B[j, n] > 0)`` when
``transpose=False`` or ``result[j, n] = sum_i A[i, j] * (B[i, n] > 0)`` when
``transpose=True``, where only active (nonzero) events in the dense matrix
``B`` contribute to the output. This is efficient for processing batches of
spike events in neural network simulations.

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``binary_coomm_p.available_backends(platform)``,
and the default backend can be configured with ``binary_coomm_p.set_default(platform, backend)``.

See Also
--------
binary_coomm : High-level user-facing function wrapper.
"""
)
binary_coomm_p.def_numba_kernel(_coomm_numba_kernel)
binary_coomm_p.def_warp_kernel(_coomm_warp_kernel)
binary_coomm_p.def_pallas_kernel('gpu', _coomm_pallas_gpu_kernel)
binary_coomm_p.def_pallas_kernel('tpu', _coomm_pallas_tpu_kernel)
binary_coomm_p.def_jvp_rule2(_coomm_jvp_left, None, None, _coomm_jvp_right)
binary_coomm_p.def_transpose_rule(_coomm_transpose_rule)
binary_coomm_p.def_batching_rule(_coomm_batching)
binary_coomm_p.def_call(binary_coomm_p_call)
binary_coomm_p.def_tags('coo', 'binary')
binary_coomm_p.def_benchmark_data(_coomm_benchmark_data)
